import socket
from ModelBuilder import ModelBuilder
from xml.dom import minidom
from ZPublisher import HTTPRequest
import AccessControl
from conga_constants import *
from ricci_bridge import *
from ricci_communicator import *
from string import lower
import time
import Products.ManagedSystem
from Products.Archetypes.utils import make_uuid
from Ip import Ip
from Clusterfs import Clusterfs
from Fs import Fs
from NFSClient import NFSClient
from NFSExport import NFSExport
from Netfs import Netfs
from Vm import Vm
from Script import Script
from Samba import Samba
from clusterOS import resolveOSType
from GeneralError import GeneralError
from UnknownClusterError import UnknownClusterError
from homebase_adapters import nodeUnauth, nodeAuth, manageCluster, createClusterSystems, havePermCreateCluster, setNodeFlag, delNodeFlag, userAuthenticated, getStorageNode, getClusterNode
from LuciSyslog import LuciSyslogError, LuciSyslog

#Policy for showing the cluster chooser menu:
#1) If there are no clusters in the ManagedClusterSystems
#folder, then only the admin user may see this menu, and
#the configure option should not be displayed.
#2)If there are clusters in the ManagedClusterSystems,
#then only display chooser if the current user has
#permissions on at least one. If the user is admin, show ALL clusters

CLUSTER_FOLDER_PATH = '/luci/systems/cluster/'

try:
	luci_log = LuciSyslog()
except LuciSyslogError, e:
	pass

def validateClusterNodes(request, sessionData, clusterName, numStorage):
	nodeList = list()
	nodeHash = {}
	rnodeHash = {}
	oldNodeHash = {}
	oldRnodeHash = {}
	requestResults = {}
	errors = list()

	if sessionData and 'requestResults' in sessionData:
		requestResults = sessionData['requestResults']
		if 'nodeHash' in requestResults:
			oldNodeHash = requestResults['nodeHash']
		if 'rnodeHash' in requestResults:
			oldRnodeHash = requestResults['rnodeHash']

	i = 0
	while i < numStorage:
		try:
			sysData = request.form['__SYSTEM' + str(i)]
			if not sysData or sysData[0] == '':
				raise

			if len(sysData) < 2 or sysData[1] == '':
				errors.append('No password was specified for host \"' + sysData[0] + '\"')
				raise
		except:
			i += 1
			continue

		if len(sysData) > 1:
			node = nodeAuth(None, sysData[0], sysData[1])

			if oldRnodeHash and node['ricci_host'] in oldRnodeHash:
				oldNode = oldRnodeHash[node['ricci_host']]
			elif oldNodeHash and node['host'] in nodeHash:
				oldNode = oldNodeHash[node['host']]
			else:
				oldNode = None

			if 'errors' in node:
				errors.append(node['errors'])
				node['errors'] = True

			if node['host'] in nodeHash or node['ricci_host'] in rnodeHash:
				node['errors'] = True
				errors.append('You added the node \"' + node['host'] + '\" more than once')
			else:
				if oldNode and 'prev_auth' in oldNode:
					node['prev_auth'] = oldNode['prev_auth']

				nodeHash[node['host']] = node
				rnodeHash[node['ricci_host']] = node
				nodeList.append(node)
		i += 1

	sfn = lambda x, y: \
		x['cur_auth'] - y['cur_auth'] or (('errors' in y) - ('errors' in x))
	nodeList.sort(sfn)

	dfn = lambda x: not 'cur_auth' in x or x['cur_auth'] != True
	cluster_properties = {
		'clusterName': clusterName,
		'nodeList': nodeList,
		'nodeHash': nodeHash,
		'rnodeHash': rnodeHash,
		'isComplete': len(errors) < 1 and len(filter(dfn, nodeList)) == 0
	}

	return [errors, cluster_properties]


def validateCreateCluster(self, request):
	errors = list()
	messages = list()
	requestResults = {}

	if not havePermCreateCluster(self):
		return (False, {'errors': ['You do not have sufficient rights to create a cluster.']})

	try:
	 	sessionData = request.SESSION.get('checkRet')
	except:
		sessionData = None

	if not 'clusterName' in request.form or not request.form['clusterName']:
		return (False, {'errors': [ 'No cluster name was specified.' ]})
	clusterName = request.form['clusterName']

	try:
		numStorage = int(request.form['numStorage'])
	except:
		return (False, { 'errors': ['Unknown number of systems entered'], 'requestResults': requestResults })

	if numStorage < 1:
		return (False, { 'errors': ['A cluster must contain at least one node'], 'requestResults': requestResults })

	ret = validateClusterNodes(request, sessionData, clusterName, numStorage)
	errors.extend(ret[0])
	cluster_properties = ret[1]

	rhn_dl = 1
	try:
		rhn_dls = request.form['rhn_dl'].strip().lower()
		if rhn_dls != '1' and rhn_dls != 'true':
			rhn_dl = 0
	except:
		rhn_dl = 0

	enable_storage = 0
	try:
		enable_storage_str = request.form['enable_storage'].strip().lower()
		if enable_storage_str:
			enable_storage = 1
	except:
		enable_storage = 0

	try:
		nodeList = cluster_properties['nodeList']
		if len(nodeList) < 1:
			raise
	except:
		errors.append('A cluster must contain at least one node')

	cluster_os = None
	try:
		cluster_os = nodeList[0]['os']
		if not cluster_os:
			raise KeyError('OS for ' + nodeList[0]['host'] + ' is blank')
	except KeyError, e:
		cluster_properties['isComplete'] = False
		errors.append('Unable to identify the operating system running on the first cluster node: ' + str(e))

	if cluster_properties['isComplete'] != True:
		nodeUnauth(nodeList)
		return (False, {'errors': errors, 'requestResults':cluster_properties })
	else:
		try:
			if len(filter(lambda x: x['os'] != cluster_os, nodeList[1:])) > 0:
				raise Exception('different operating systems were detected.')
		except:
			cluster_properties['isComplete'] = False
			errors.append('Cluster nodes must be running compatible operating systems.')

	if cluster_properties['isComplete'] == True:
		batchNode = createClusterBatch(cluster_os,
						clusterName,
						clusterName,
						map(lambda x: x['ricci_host'], nodeList),
						True,
						True,
						enable_storage,
						False,
						rhn_dl)

		if not batchNode:
			nodeUnauth(nodeList)
			cluster_properties['isComplete'] = False
			errors.append('Unable to generate cluster creation ricci command')
			return (False, {'errors': errors, 'requestResults':cluster_properties })

		error = manageCluster(self, clusterName, nodeList)
		if error:
			nodeUnauth(nodeList)
			cluster_properties['isComplete'] = False
			errors.append(error)
			return (False, {'errors': errors, 'requestResults':cluster_properties })

		batch_id_map = {}
		rc = None
		for i in nodeList:
			success = True
			try:
				rc = RicciCommunicator(i['ricci_host'])
			except RicciError, e:
				luci_log.debug('Unable to connect to the ricci agent on %s: %s'\
					% (i['ricci_host'], str(e)))
				success = False
			except:
				success = False

			if success == True:
				try:
					resultNode = rc.process_batch(batchNode, async=True)
					batch_id_map[i['ricci_host']] = resultNode.getAttribute('batch_id')
				except:
					success = False

			if not success:
				nodeUnauth(nodeList)
				cluster_properties['isComplete'] = False
				errors.append('An error occurred while attempting to add cluster node \"' + i['ricci_host'] + '\"')
				return (False, {'errors': errors, 'requestResults':cluster_properties })
		buildClusterCreateFlags(self, batch_id_map, clusterName)

	messages.append('Creation of cluster \"' + clusterName + '\" has begun')
	return (True, {'errors': errors, 'messages': messages })

def buildClusterCreateFlags(self, batch_map, clusterName):
  path = str(CLUSTER_FOLDER_PATH + clusterName)
  clusterfolder = self.restrictedTraverse(path)
  for key in batch_map.keys():
    key = str(key)
    id = batch_map[key]
    batch_id = str(id)
    objname = str(key + "____flag") #This suffix needed to avoid name collision
    clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
    #now designate this new object properly
    objpath = str(path + "/" + objname)
    flag = self.restrictedTraverse(objpath)
    #flag[BATCH_ID] = batch_id
    #flag[TASKTYPE] = CLUSTER_ADD
    #flag[FLAG_DESC] = "Creating node " + key + " for cluster " + clusterName
    flag.manage_addProperty(BATCH_ID,batch_id, "string")
    flag.manage_addProperty(TASKTYPE,CLUSTER_ADD, "string")
    flag.manage_addProperty(FLAG_DESC,"Creating node " + key + " for cluster " + clusterName, "string")
    flag.manage_addProperty(LAST_STATUS, 0, "int")

def validateAddClusterNode(self, request):
	errors = list()
	messages = list()
	requestResults = {}

	try:
	 	sessionData = request.SESSION.get('checkRet')
	except:
		sessionData = None

	if 'clusterName' in request.form:
		clusterName = str(request.form['clusterName'])
	else:
		return (False, {'errors': [ 'Cluster name is missing'], 'requestResults': requestResults })

	rhn_dl = 1
	try:
		rhn_dls = request.form['rhn_dl'].strip().lower()
		if rhn_dls != '1' and rhn_dls != 'true':
			rhn_dl = 0
	except:
		rhn_dl = 0

	enable_storage = 0
	try:
		enable_storages = request.form['enable_storage'].strip().lower()
		if enable_storages:
			enable_storage = 1
	except:
		enable_storage = 0

	try:
		numStorage = int(request.form['numStorage'])
		if numStorage < 1:
			raise
	except:
		errors.append('You must specify at least one node to add to the cluster')
		return (False, {'errors': [ errors ], 'requestResults': requestResults })

	ret = validateClusterNodes(request, sessionData, clusterName, numStorage)
	errors.extend(ret[0])
	cluster_properties = ret[1]

	try:
		nodeList = cluster_properties['nodeList']
		if len(nodeList) < 1:
			raise
	except:
		errors.append('You must specify at least one valid node to add to the cluster')

	try:
		clusterObj = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		cluster_os = clusterObj.manage_getProperty('cluster_os')
		if not cluster_os:
			luci_log.debug('The cluster OS property is missing for cluster ' + clusterName)
			raise Exception, 'no cluster OS was found.'
		try:
			if len(filter(lambda x: x['os'] != cluster_os, nodeList)) > 0:
				raise Exception, 'different operating systems were detected.'
		except:
			nodeUnauth(nodeList)
			cluster_properties['isComplete'] = False
			errors.append('Cluster nodes must be running compatible operating systems.')
	except:
		nodeUnauth(nodeList)
		cluster_properties['isComplete'] = False
		errors.append('Unable to determine the cluster OS for the ' + clusterName + ' cluster.')

	if not cluster_properties['isComplete']:
		return (False, {'errors': errors, 'requestResults': cluster_properties})

	i = 0
	while i < len(nodeList):
		clunode = nodeList[i]
		try:
			batchNode = addClusterNodeBatch(clusterName,
							True,
							True,
							enable_storage,
							False,
							rhn_dl)
			if not batchNode:
				raise
			del nodeList[i]
		except:
			clunode['errors'] = True
			nodeUnauth(nodeList)
			cluster_properties['isComplete'] = False
			errors.append('Unable to initiate node creation for host \"' + clunode['ricci_host'] + '\"')

	if not cluster_properties['isComplete']:
		return (False, {'errors': errors, 'requestResults': cluster_properties})

	error = createClusterSystems(self, clusterName, nodeList)
	if error:
		nodeUnauth(nodeList)
		cluster_properties['isComplete'] = False
		errors.append(error)
		return (False, {'errors': errors, 'requestResults': cluster_properties})

	batch_id_map = {}
	for i in nodeList:
		clunode = nodeList[i]
		success = True
		try:
			rc = RicciCommunicator(clunode['ricci_host'])
		except:
			luci_log.info('Unable to connect to the ricci daemon on host ' + clunode['ricci_host'])
			success = False

		if success:
			try:
				resultNode = rc.process_batch(batchNode, async=True)
				batch_id_map[clunode['ricci_host']] = resultNode.getAttribute('batch_id')
			except:
				success = False

		if not success:
			nodeUnauth(nodeList)
			cluster_properties['isComplete'] = False
			errors.append('An error occurred while attempting to add cluster node \"' + clunode['ricci_host'] + '\"')
			return (False, {'errors': errors, 'requestResults': cluster_properties})

			messages.append('Cluster join initiated for host \"' + clunode['ricci_host'] + '\"')

	buildClusterCreateFlags(self, batch_id_map, clusterName)
	return (True, {'errors': errors, 'messages': messages})

def validateServiceAdd(self, request):
	try:
		form_xml = request['form_xml']
		if not form_xml:
			raise KeyError('form_xml must not be blank')
	except KeyError, e:
		return (False, {'errors': ['No resource data was supplied for this service.']})

	try:
		doc = minidom.parseString(form_xml)
		forms = doc.getElementsByTagName('form')
		if len(forms) < 1:
			raise
	except:
		return (False, {'errors': ['The resource data submitted for this service is not properly formed.']})

	form_hash = {}
	form_hash['toplevel'] = {'form': None, 'kids': [] }
	for i in forms:
		form_id = i.getAttribute('id')
		form_parent = i.getAttribute('parent')
		if not form_id or not form_parent:
			continue
		ielems = i.getElementsByTagName('input')
		if not ielems or len(ielems) < 1:
			continue
		if not form_id in form_hash:
			form_hash[form_id] = {'form': i, 'kids': []}
		elif not form_hash[form_id]['form']:
			form_hash[form_id]['form'] = i
		if not form_parent in form_hash:
			form_hash[form_parent] = {'form': None, 'kids': []}
		form_hash[form_parent]['kids'].append(form_id)
		dummy_form = {}
		for i in ielems:
			try:
				type = str(i.getAttribute('type'))
			except:
				continue
			if not type or type == 'button':
				continue
			try:
				dummy_form[str(i.getAttribute('name'))] = str(i.getAttribute('value'))
			except:
				pass

		try:
			res_type = dummy_form['type'].strip()
			if not res_type or not res_type in resourceAddHandler:
				raise
		except:
			return (False, {'errors': ['An invalid resource type was specified: ' + res_type]})

		try:
			resObj = resourceAddHandler[res_type](request, dummy_form)
		except:
			luci_log('res type %d is invalid' % res_type)
			resObj = None

		if resObj is None:
			return (False, {'errors': ['An error occurred while adding ' + res_type]})
		form_hash[form_id]['obj'] = resObj
			
	return (True, {'messages': ['This service has been updated.']})

def validateResourceAdd(self, request):
	try:
		res_type = request.form['type'].strip()
		if not res_type:
			raise KeyError, 'type is blank'
	except Exception, e:
		luci_log.debug_verbose('resourceAdd: type is blank')
		return (False, {'errors': ['No resource type was given.']})

	errors = list()
	try:
		res = resourceAddHandler[res_type](request)
		if res is None or res[0] is None or res[1] is None:
			if res and res[2]:
				errors.extend(res[2])
			raise Exception, 'An error occurred while adding this resource'
		modelb = res[1]
		newres = res[0]
		addResource(self, request, modelb, newres)
	except Exception, e:
		if len(errors) < 1:
			errors.append('An error occurred while adding this resource')
		luci_log.debug_verbose('resource error: %s' % str(e))
		return (False, {'errors': errors})

	return (True, {'messages': ['Resource added successfully']})
	
## Cluster properties form validation routines

def validateMCastConfig(self, form):
	try:
		mcast_val = form['mcast'].strip().lower()
		if mcast_val != 'true' and mcast_val != 'false':
			raise KeyError(mcast_val)
		if mcast_val == 'true':
			mcast_val = 1
		else:
			mcast_val = 0
	except KeyError, e:
		return (False, {'errors': ['An invalid multicast selection was made.']})

	if not mcast_val:
		return (True, {'messages': ['Changes accepted. - FILL ME IN']})

	try:
		addr_str = form['mcast_addr'].strip()
		socket.inet_pton(socket.AF_INET, addr_str)
	except KeyError, e:
		return (False, {'errors': ['No multicast address was given']})
	except socket.error, e:
		try:
			socket.inet_pton(socket.AF_INET6, addr_str)
		except socket.error, e6:
			return (False, {'errors': ['An invalid multicast address was given: ' + e]})

	return (True, {'messages': ['Changes accepted. - FILL ME IN']})

def validateQDiskConfig(self, form):
	errors = list()

	try:
		qdisk_val = form['quorumd'].strip().lower()
		if qdisk_val != 'true' and qdisk_val != 'false':
			raise KeyError(qdisk_val)
		if qdisk_val == 'true':
			qdisk_val = 1
		else:
			qdisk_val = 0
	except KeyError, e:
		return (False, {'errors': ['An invalid quorum partition selection was made.']})

	if not qdisk_val:
		return (True, {'messages': ['Changes accepted. - FILL ME IN']})

	try:
		interval = int(form['interval'])
		if interval < 0:
			raise ValueError('Interval must be 0 or greater.')
	except KeyError, e:
		errors.append('No Interval value was given.')
	except ValueError, e:
		errors.append('An invalid Interval value was given: ' + e)

	try:
		votes = int(form['votes'])
		if votes < 1:
			raise ValueError('Votes must be greater than 0')
	except KeyError, e:
		errors.append('No Votes value was given.')
	except ValueError, e:
		errors.append('An invalid Votes value was given: ' + e)

	try:
		tko = int(form['tko'])
		if tko < 0:
			raise ValueError('TKO must be 0 or greater')
	except KeyError, e:
		errors.append('No TKO value was given.')
	except ValueError, e:
		errors.append('An invalid TKO value was given: ' + e)

	try:
		min_score = int(form['min_score'])
		if min_score < 1:
			raise ValueError('Minimum Score must be greater than 0')
	except KeyError, e:
		errors.append('No Minimum Score value was given.')
	except ValueError, e:
		errors.append('An invalid Minimum Score value was given: ' + e)

	try:
		device = form['device'].strip()
		if not device:
			raise KeyError('device')
	except KeyError, e:
		errors.append('No Device value was given.')

	try:
		label = form['label'].strip()
		if not label:
			raise KeyError('label')
	except KeyError, e:
		errors.append('No Label value was given.')

	num_heuristics = 0
	try:
		num_heuristics = int(form['num_heuristics'])
		if num_heuristics < 0:
			raise ValueError(form['num_heuristics'])
		if num_heuristics == 0:
			num_heuristics = 1
	except KeyError, e:
		errors.append('No number of heuristics was given.')
	except ValueError, e:
		errors.append('An invalid number of heuristics was given: ' + e)

	heuristics = list()
	for i in xrange(num_heuristics):
		prefix = 'heuristic' + str(i) + ':'
		try:
			hname = form[prefix + 'hname'].strip()
			if not hname:
				raise KeyError(prefix + 'hname')
		except KeyError, e:
			if ((not prefix + 'hpath' in form or not form['hpath'].strip()) and
				(not prefix + 'hint' in form or not form['hint'].strip()) and
				(not prefix + 'hscore' in form or not form['hscore'].strip())):
				# The row is blank; ignore it.
				continue
			errors.append('No heuristic name was given for heuristic #' + str(i + 1))

		try:
			hpath = form[prefix + 'hpath']
		except KeyError, e:
			errors.append('No heuristic path was given for heuristic #' + str(i + 1))

		try:
			hint = int(form[prefix + 'hint'])
			if hint < 1:
				raise ValueError('Heuristic interval values must be greater than 0.')
		except KeyError, e:
			errors.append('No heuristic interval was given for heuristic #' + str(i + 1))
		except ValueError, e:
			errors.append('An invalid heuristic interval was given for heuristic #' + str(i + 1) + ': ' + e)

		try:
			hscore = int(form[prefix + 'score'])
			if hscore < 1:
				raise ValueError('Heuristic scores must be greater than 0.')
		except KeyError, e:
			errors.append('No heuristic score was given for heuristic #' + str(i + 1))
		except ValueError, e:
			errors.append('An invalid heuristic score was given for heuristic #' + str(i + 1) + ': ' + e)
		heuristics.append([ hname, hpath, hint, hscore ])

	if len(errors) > 0:
		return (False, {'errors': errors })
	return (True, {'messages': ['Changes accepted. - FILL ME IN']})

def validateGeneralConfig(self, form):
	errors = list()

	try:
		cluster_name = form['cluname'].strip()
		if not cluster_name:
			raise KeyError('cluname')
	except KeyError, e:
		errors.append('No cluster name was given.')

	try:
		version_num = int(form['cfgver'])
		if version_num < 0:
			raise ValueError('configuration version numbers must be 0 or greater.')
	except KeyError, e:
		errors.append('No cluster configuration version was given.')
	except ValueError, e:
		errors.append('An invalid configuration version was given: ' + e)

	if len(errors) > 0:
		return (False, {'errors': errors})

	return (True, {'messages': ['Changes accepted. - FILL ME IN']})

def validateFenceConfig(self, form):
	errors = list()

	try:
		post_fail_delay = int(form['post_fail_delay'])
		if post_fail_delay < 0:
			raise ValueError('post fail delay values must be 0 or greater.')
	except KeyError, e:
		errors.append('No post fail delay was given.')
	except ValueError, e:
		errors.append('Invalid post fail delay: ' + e)

	try:
		post_join_delay = int(form['post_join_delay'])
		if post_join_delay < 0:
			raise ValueError('post join delay values must be 0 or greater.')
	except KeyError, e:
		errors.append('No post join delay was given.')
	except ValueError, e:
		errors.append('Invalid post join delay: ' + e)

	if len(errors) > 0:
		return (False, {'errors': errors })

	return (True, {'messages': ['Changes accepted. - FILL ME IN']})

configFormValidators = {
	'general': validateGeneralConfig,
	'mcast': validateMCastConfig,
	'fence': validateFenceConfig,
	'qdisk': validateQDiskConfig
}

def validateConfigCluster(self, request):
	errors = list()
	messages = list()

	if not 'form' in request:
		return (False, {'errors': ['No form was submitted.']})
	if not 'configtype' in request.form:
		return (False, {'errors': ['No configuration type was submitted.']})
	if not request.form['configtype'] in configFormValidators:
		return (False, {'errors': ['An invalid configuration type was submitted.']})

	val = configFormValidators[request.form['configtype']]
	ret = val(self, request.form)

	retcode = ret[0]
	if 'errors' in ret[1]:
		errors.extend(ret[1]['errors'])
	if 'messages' in ret[1]:
		messages.extend(ret[1]['messages'])

	if len(errors) < 1:
		messages.append('The cluster properties have been updated.')

	return (retcode, {'errors': errors, 'messages': messages})

def validateFenceAdd(self, request):
	return (True, {})

def validateFenceEdit(self, request):
	return (True, {})

formValidators = {
	6: validateCreateCluster,
	7: validateConfigCluster,
	15: validateAddClusterNode,
	21: validateServiceAdd,
	24: validateServiceAdd,
	31: validateResourceAdd,
	33: validateResourceAdd,
	51: validateFenceAdd,
	50: validateFenceEdit,
}

def validatePost(self, request):
	pagetype = int(request.form['pagetype'])
	if not pagetype in formValidators:
		return None
	else:
		return formValidators[pagetype](self, request)


def createCluChooser(self, request, systems):
  dummynode = {}

  if request.REQUEST_METHOD == 'POST':
    ret = validatePost(self, request)
    try:
      request.SESSION.set('checkRet', ret[1])
    except:
      request.SESSION.set('checkRet', {})
  else:
    try: request.SESSION.set('checkRet', {})
    except: pass

  #First, see if a cluster is chosen, then
  #check that the current user can access that system
  cname = None
  try:
    cname = request[CLUNAME]
  except KeyError, e:
    cname = ""

  try:
    url = request['URL']
  except KeyError, e:
    url = "."

  try:
    pagetype = request[PAGETYPE]
  except KeyError, e:
    pagetype = '3'


  cldata = {}
  cldata['Title'] = "Cluster List"
  cldata['cfg_type'] = "clusters"
  cldata['absolute_url'] = url + "?pagetype=" + CLUSTERLIST
  cldata['Description'] = "Clusters available for configuration"
  if pagetype == CLUSTERLIST:
    cldata['currentItem'] = True
  else:
    cldata['currentItem'] = False

  UserHasPerms = havePermCreateCluster(self)
  if UserHasPerms:
    cladd = {}
    cladd['Title'] = "Create a New Cluster"
    cladd['cfg_type'] = "clusteradd"
    cladd['absolute_url'] = url + "?pagetype=" + CLUSTER_ADD
    cladd['Description'] = "Create a Cluster"
    if pagetype == CLUSTER_ADD:
      cladd['currentItem'] = True
    else:
      cladd['currentItem'] = False

  clcfg = {}
  clcfg['Title'] = "Configure"
  clcfg['cfg_type'] = "clustercfg"
  clcfg['absolute_url'] = url + "?pagetype=" + CLUSTERS
  clcfg['Description'] = "Configure a cluster"
  if pagetype == CLUSTERS:
    clcfg['currentItem'] = True
  else:
    clcfg['currentItem'] = False

  #test...
  #clcfg['show_children'] = True
  #Add all cluster type pages here:
  if pagetype == CLUSTER or pagetype == CLUSTER_CONFIG:
    clcfg['show_children'] = True
  else:
    clcfg['show_children'] = False

  #loop through all clusters
  syslist= list()
  for system in systems:
    clsys = {}
    clsys['Title'] = system[0]
    clsys['cfg_type'] = "cluster"
    clsys['absolute_url'] = url + "?pagetype=" + CLUSTER + "&clustername=" + system[0]
    clsys['Description'] = "Configure this cluster"

    if pagetype == CLUSTER or pagetype == CLUSTER_CONFIG:
      if cname == system[0]:
        clsys['currentItem'] = True
      else:
        clsys['currentItem'] = False
    else:
      clsys['currentItem'] = False
    syslist.append(clsys)

  clcfg['children'] = syslist

  mylist = list()
  mylist.append(cldata)
  if UserHasPerms:
    mylist.append(cladd)
  mylist.append(clcfg)
  dummynode['children'] = mylist

  return dummynode


def createCluConfigTree(self, request, model):
  dummynode = {}

  #There should be a positive page type
  try:
    pagetype = request[PAGETYPE]
  except KeyError, e:
    pagetype = '3'

  try:
    url = request['URL']
  except KeyError, e:
    url = "."

  #The only way this method can run is if there exists
  #a clustername query var
  cluname = request['clustername']


  nd = {}
  nd['Title'] = "Nodes"
  nd['cfg_type'] = "nodes"
  nd['absolute_url'] = url + "?pagetype=" + NODES + "&clustername=" + cluname
  nd['Description'] = "Node configuration for this cluster"
  if pagetype == NODES or pagetype == NODE_GRID or pagetype == NODE_LIST or pagetype == NODE_CONFIG or pagetype == NODE_ADD or pagetype == NODE:
    nd['show_children'] = True
  else:
    nd['show_children'] = False
  if pagetype == "0":
    nd['show_children'] = False

  if pagetype == NODES:
    nd['currentItem'] = True
  else:
    nd['currentItem'] = False


  ndadd = {}
  ndadd['Title'] = "Add a Node"
  ndadd['cfg_type'] = "nodeadd"
  ndadd['absolute_url'] = url + "?pagetype=" + NODE_ADD + "&clustername=" + cluname
  ndadd['Description'] = "Add a node to this cluster"
  if pagetype == NODE_ADD:
    ndadd['currentItem'] = True
  else:
    ndadd['currentItem'] = False

  ndcfg = {}
  ndcfg['Title'] = "Configure"
  ndcfg['cfg_type'] = "nodecfg"
  ndcfg['absolute_url'] = url + "?pagetype=" + NODE_CONFIG + "&clustername=" + cluname
  ndcfg['Description'] = "Configure cluster nodes"
  if pagetype == NODE_CONFIG or pagetype == NODE or pagetype == NODES or pagetype == NODE_LIST or pagetype == NODE_GRID or pagetype == NODE_ADD:
    ndcfg['show_children'] = True
  else:
    ndcfg['show_children'] = False
  if pagetype == NODE_CONFIG:
    ndcfg['currentItem'] = True
  else:
    ndcfg['currentItem'] = False

  nodes = model.getNodes()
  nodenames = list()
  for node in nodes:
    nodenames.append(node.getName())

  cfgablenodes = list()
  for nodename in nodenames:
    cfg = {}
    cfg['Title'] = nodename
    cfg['cfg_type'] = "node"
    cfg['absolute_url'] = url + "?pagetype=" + NODE + "&nodename=" + nodename + "&clustername=" + cluname
    cfg['Description'] = "Configure this cluster node"
    if pagetype == NODE:
      try:
        nname = request['nodename']
      except KeyError, e:
        nname = ""
      if nodename == nname:
        cfg['currentItem'] = True
      else:
        cfg['currentItem'] = False
    else:
      cfg['currentItem'] = False

    cfgablenodes.append(cfg)

  #Now add nodename structs as children of the config element
  ndcfg['children'] = cfgablenodes

  ndkids = list()
  ndkids.append(ndadd)
  ndkids.append(ndcfg)

  nd['children'] = ndkids

  ##################################################################
  sv = {}
  sv['Title'] = "Services"
  sv['cfg_type'] = "services"
  sv['absolute_url'] = url + "?pagetype=" + SERVICES + "&clustername=" + cluname
  sv['Description'] = "Service configuration for this cluster"
  if pagetype == SERVICES or pagetype == SERVICE_CONFIG or pagetype == SERVICE_ADD or pagetype == SERVICE:
    sv['show_children'] = True
  else:
    sv['show_children'] = False
  if pagetype == SERVICES:
    sv['currentItem'] = True
  else:
    sv['currentItem'] = False

  svadd = {}
  svadd['Title'] = "Add a Service"
  svadd['cfg_type'] = "serviceadd"
  svadd['absolute_url'] = url + "?pagetype=" + SERVICE_ADD + "&clustername=" + cluname
  svadd['Description'] = "Add a Service to this cluster"
  if pagetype == SERVICE_ADD:
    svadd['currentItem'] = True
  else:
    svadd['currentItem'] = False

  if model.getIsVirtualized() == True:
    vmadd = {}
    vmadd['Title'] = "Add a Virtual Service"
    vmadd['cfg_type'] = "xenvmadd"
    vmadd['absolute_url'] = url + "?pagetype=" + XENVM_ADD + "&clustername=" + cluname
    vmadd['Description'] = "Add a Virtual Service to this cluster"
    if pagetype == XENVM_ADD:
      vmadd['currentItem'] = True
    else:
      vmadd['currentItem'] = False

  svcfg = {}
  svcfg['Title'] = "Configure a Service"
  svcfg['cfg_type'] = "servicecfg"
  svcfg['absolute_url'] = url + "?pagetype=" + SERVICE_CONFIG + "&clustername=" + cluname
  svcfg['Description'] = "Configure a Service for this cluster"
  if pagetype == SERVICE_CONFIG or pagetype == SERVICE:
    svcfg['show_children'] = True
  else:
    svcfg['show_children'] = False
  if pagetype == SERVICE_CONFIG:
    svcfg['currentItem'] = True
  else:
    svcfg['currentItem'] = False

  services = model.getServices()
  xenvms = model.getXENVMs()
  serviceable = list()
  for service in services:
    servicename = service.getName()
    svc = {}
    svc['Title'] = servicename
    svc['cfg_type'] = "service"
    svc['absolute_url'] = url + "?pagetype=" + SERVICE + "&servicename=" + servicename + "&clustername=" + cluname
    svc['Description'] = "Configure this service"
    if pagetype == SERVICE:
      try:
        sname = request['servicename']
      except KeyError, e:
        sname = ""
      if servicename == sname:
        svc['currentItem'] = True
      else:
        svc['currentItem'] = False
    else:
      svc['currentItem'] = False

    serviceable.append(svc)

  for xenvm in xenvms:
    xenname = xenvm.getName()
    svc = {}
    svc['Title'] = xenname
    svc['cfg_type'] = "xenvm"
    svc['absolute_url'] = url + "?pagetype=" + XENVM_CONFIG + "&servicename=" + xenname + "&clustername=" + cluname
    svc['Description'] = "Configure this Virtual Service"
    if pagetype == XENVM_CONFIG:
      try:
        xname = request['servicename']
      except KeyError, e:
        xname = ""
      if xenname == xname:
        svc['currentItem'] = True
      else:
        svc['currentItem'] = False
    else:
      svc['currentItem'] = False

    serviceable.append(svc)

  svcfg['children'] = serviceable



  kids = list()
  kids.append(svadd)
  if model.getIsVirtualized() == True:
    kids.append(vmadd)
  kids.append(svcfg)
  sv['children'] = kids
#############################################################
  rv = {}
  rv['Title'] = "Resources"
  rv['cfg_type'] = "resources"
  rv['absolute_url'] = url + "?pagetype=" + RESOURCES + "&clustername=" + cluname
  rv['Description'] = "Resource configuration for this cluster"
  if pagetype == RESOURCES or pagetype == RESOURCE_CONFIG or pagetype == RESOURCE_ADD or pagetype == RESOURCE:
    rv['show_children'] = True
  else:
    rv['show_children'] = False
  if pagetype == RESOURCES:
    rv['currentItem'] = True
  else:
    rv['currentItem'] = False

  rvadd = {}
  rvadd['Title'] = "Add a Resource"
  rvadd['cfg_type'] = "resourceadd"
  rvadd['absolute_url'] = url + "?pagetype=" + RESOURCE_ADD + "&clustername=" + cluname
  rvadd['Description'] = "Add a Resource to this cluster"
  if pagetype == RESOURCE_ADD:
    rvadd['currentItem'] = True
  else:
    rvadd['currentItem'] = False

  rvcfg = {}
  rvcfg['Title'] = "Configure a Resource"
  rvcfg['cfg_type'] = "resourcecfg"
  rvcfg['absolute_url'] = url + "?pagetype=" + RESOURCE_CONFIG + "&clustername=" + cluname
  rvcfg['Description'] = "Configure a Resource for this cluster"
  if pagetype == RESOURCE_CONFIG or pagetype == RESOURCE:
    rvcfg['show_children'] = True
  else:
    rvcfg['show_children'] = False
  if pagetype == RESOURCE_CONFIG:
    rvcfg['currentItem'] = True
  else:
    rvcfg['currentItem'] = False

  resources = model.getResources()
  resourceable = list()
  for resource in resources:
    resourcename = resource.getName()
    rvc = {}
    rvc['Title'] = resourcename
    rvc['cfg_type'] = "resource"
    rvc['absolute_url'] = url + "?pagetype=" + RESOURCE + "&resourcename=" + resourcename + "&clustername=" + cluname
    rvc['Description'] = "Configure this resource"
    if pagetype == RESOURCE:
      try:
        rname = request['resourcename']
      except KeyError, e:
        rname = ""
      if resourcename == rname:
        rvc['currentItem'] = True
      else:
        rvc['currentItem'] = False
    else:
      rvc['currentItem'] = False

    resourceable.append(rvc)
  rvcfg['children'] = resourceable



  kids = list()
  kids.append(rvadd)
  kids.append(rvcfg)
  rv['children'] = kids
 #################################################################
  fd = {}
  fd['Title'] = "Failover Domains"
  fd['cfg_type'] = "failoverdomains"
  fd['absolute_url'] = url + "?pagetype=" + FDOMS + "&clustername=" + cluname
  fd['Description'] = "Failover domain configuration for this cluster"
  if pagetype == FDOMS or pagetype == FDOM_CONFIG or pagetype == FDOM_ADD or pagetype == FDOM:
    fd['show_children'] = True
  else:
    fd['show_children'] = False
  if pagetype == FDOMS:
    fd['currentItem'] = True
  else:
    fd['currentItem'] = False

  fdadd = {}
  fdadd['Title'] = "Add a Failover Domain"
  fdadd['cfg_type'] = "failoverdomainadd"
  fdadd['absolute_url'] = url + "?pagetype=" + FDOM_ADD + "&clustername=" + cluname
  fdadd['Description'] = "Add a Failover Domain to this cluster"
  if pagetype == FDOM_ADD:
    fdadd['currentItem'] = True
  else:
    fdadd['currentItem'] = False

  fdcfg = {}
  fdcfg['Title'] = "Configure a Failover Domain"
  fdcfg['cfg_type'] = "failoverdomaincfg"
  fdcfg['absolute_url'] = url + "?pagetype=" + FDOM_CONFIG + "&clustername=" + cluname
  fdcfg['Description'] = "Configure a Failover Domain for this cluster"
  if pagetype == FDOM_CONFIG or pagetype == FDOM:
    fdcfg['show_children'] = True
  else:
    fdcfg['show_children'] = False
  if pagetype == FDOM_CONFIG:
    fdcfg['currentItem'] = True
  else:
    fdcfg['currentItem'] = False

  fdoms = model.getFailoverDomains()
  fdomable = list()
  for fdom in fdoms:
    fdomname = fdom.getName()
    fdc = {}
    fdc['Title'] = fdomname
    fdc['cfg_type'] = "fdom"
    fdc['absolute_url'] = url + "?pagetype=" + FDOM + "&fdomname=" + fdomname + "&clustername=" + cluname
    fdc['Description'] = "Configure this Failover Domain"
    if pagetype == FDOM:
      try:
        fname = request['fdomname']
      except KeyError, e:
        fname = ""
      if fdomname == fname:
        fdc['currentItem'] = True
      else:
        fdc['currentItem'] = False
    else:
      fdc['currentItem'] = False

    fdomable.append(fdc)
  fdcfg['children'] = fdomable



  kids = list()
  kids.append(fdadd)
  kids.append(fdcfg)
  fd['children'] = kids
#############################################################
  fen = {}
  fen['Title'] = "Shared Fence Devices"
  fen['cfg_type'] = "fencedevicess"
  fen['absolute_url'] = url + "?pagetype=" + FENCEDEVS + "&clustername=" + cluname
  fen['Description'] = "Fence Device configuration for this cluster"
  if pagetype == FENCEDEVS or pagetype == FENCEDEV_CONFIG or pagetype == FENCEDEV_ADD or pagetype == FENCEDEV:
    fen['show_children'] = True
  else:
    fen['show_children'] = False
  if pagetype == FENCEDEVS:
    fen['currentItem'] = True
  else:
    fen['currentItem'] = False

  fenadd = {}
  fenadd['Title'] = "Add a Fence Device"
  fenadd['cfg_type'] = "fencedeviceadd"
  fenadd['absolute_url'] = url + "?pagetype=" + FENCEDEV_ADD + "&clustername=" + cluname
  fenadd['Description'] = "Add a Fence Device to this cluster"
  if pagetype == FENCEDEV_ADD:
    fenadd['currentItem'] = True
  else:
    fenadd['currentItem'] = False

  fencfg = {}
  fencfg['Title'] = "Configure a Fence Device"
  fencfg['cfg_type'] = "fencedevicecfg"
  fencfg['absolute_url'] = url + "?pagetype=" + FENCEDEV_CONFIG + "&clustername=" + cluname
  fencfg['Description'] = "Configure a Fence Device for this cluster"
  if pagetype == FENCEDEV_CONFIG or pagetype == FENCEDEV:
    fencfg['show_children'] = True
  else:
    fencfg['show_children'] = False
  if pagetype == FENCEDEV_CONFIG:
    fencfg['currentItem'] = True
  else:
    fencfg['currentItem'] = False

  fences = model.getFenceDevices()
  fenceable = list()
  for fence in fences:
    fencename = fence.getName()
    fenc = {}
    fenc['Title'] = fencename
    fenc['cfg_type'] = "fencedevice"
    fenc['absolute_url'] = url + "?pagetype=" + FENCEDEV + "&fencename=" + fencename + "&clustername=" + cluname
    fenc['Description'] = "Configure this Fence Device"
    if pagetype == FENCEDEV:
      try:
        fenname = request['fencename']
      except KeyError, e:
        fenname = ""
      if fencename == fenname:
        fenc['currentItem'] = True
      else:
        fenc['currentItem'] = False
    else:
      fenc['currentItem'] = False

    fenceable.append(fenc)
  fencfg['children'] = fenceable



  kids = list()
  kids.append(fenadd)
  kids.append(fencfg)
  fen['children'] = kids
#############################################################

  mylist = list()
  mylist.append(nd)
  mylist.append(sv)
  mylist.append(rv)
  mylist.append(fd)
  mylist.append(fen)

  dummynode['children'] = mylist

  return dummynode


def getClusterName(self, model):
  return model.getClusterName()

def getClusterAlias(self, model):
  alias = model.getClusterAlias()
  if alias == None:
    return model.getClusterName()
  else:
    return alias

def getClusterURL(self, request, model):
  #this method will eventually extract the clustername from the request method
  return "."

def getTabs(self, req):
  ###XXX Make this method return only tabs current user can use
  portaltabs = list()
  if not userAuthenticated(self):
    return portaltabs
  selectedtab = "homebase"
  try:
    baseurl = req['URL']
    if baseurl.find("cluster") > (-1):
      selectedtab = "cluster"
    elif baseurl.find("storage") > (-1):
      selectedtab = "storage"
    else:
      selectedtab = "homebase"
  except KeyError, e:
    pass

  try:
    base2 = req['BASE2']
  except KeyError, e:
    base2 = req['HTTP_HOST'] + req['SERVER_PORT']

  htab = { 'Title':"homebase",
           'Description':"Home base for this luci server",
           'Taburl':"/luci/homebase"}
  if selectedtab == "homebase":
    htab['isSelected'] = True
  else:
    htab['isSelected'] = False


  ctab = { 'Title':"cluster",
           'Description':"Cluster configuration page",
           'Taburl':"/luci/cluster?pagetype=3"}
  if selectedtab == "cluster":
    ctab['isSelected'] = True
  else:
    ctab['isSelected'] = False

  stab = { 'Title':"storage",
           'Description':"Storage configuration page",
           'Taburl':"/luci/storage"}
  if selectedtab == "storage":
    stab['isSelected'] = True
  else:
    stab['isSelected'] = False

  portaltabs.append(htab)
  portaltabs.append(ctab)
  portaltabs.append(stab)

  return portaltabs



def check_clusters(self,clusters):
  clist = list()
  for cluster in clusters:
    if cluster_permission_check(cluster[1]):
      clist.append(cluster)

  return clist

def cluster_permission_check(cluster):
	#Does this take too long?
	try:
		sm = AccessControl.getSecurityManager()
		user = sm.getUser()
		if user.has_permission('View', cluster):
			return True
	except:
		pass
	return False

def getRicciAgent(self, clustername):
	#Check cluster permission here! return none if false
	path = str(CLUSTER_FOLDER_PATH + clustername)

	try:
		clusterfolder = self.restrictedTraverse(path)
		if not clusterfolder:
			luci_log.debug('GRA: cluster folder %s for %s is missing.' \
				% (path, clustername))
			raise Exception, 'no cluster folder at %s' % path
		nodes = clusterfolder.objectItems('Folder')
		if len(nodes) < 1:
			luci_log.debug('GRA: no cluster nodes for %s found.' % clustername)
			raise Exception, 'no cluster nodes were found at %s' % path
	except Exception, e:
		luci_log.debug('GRA: cluster folder %s for %s is missing: %s.' \
			% (path, clustername, str(e)))
		return None

	cluname = lower(clustername)

	for node in nodes:
		try:
			hostname = node[1].getId()
		except:
			try:
				hostname = node[0]
			except:
				continue

		try:
			rc = RicciCommunicator(hostname)
		except RicciError, e:
			luci_log.debug('GRA: ricci error: %s' % str(e))
			continue

		try:
			clu_info = rc.cluster_info()
		except Exception, e:
			luci_log.debug('GRA: cluster_info error: %s' % str(e))

		if cluname != lower(clu_info[0]) and cluname != lower(clu_info[1]):
			try:
				luci_log.debug('GRA: %s reports it\'s in cluster %s:%s; we expect %s' \
					 % (hostname, clu_info[0], clu_info[1], cluname))
				setNodeFlag(self, node, CLUSTER_NODE_NOT_MEMBER)
			except:
				pass
			continue

		if rc.authed():
			return rc
		try:
			setNodeFlag(node[1], CLUSTER_NODE_NEED_AUTH)
		except:
			pass

	luci_log.debug('GRA: no ricci agent could be found for cluster %s' % cluname)
	return None

def getRicciAgentForCluster(self, req):
	try:
		clustername = req['clustername']
	except KeyError, e:
		try:
			clustername = req.form['clusterName']
			if not clustername:
				raise
		except:
			luci_log.debug('no cluster name was specified in getRicciAgentForCluster')
			return None
	return getRicciAgent(self, clustername)

def getClusterStatus(self, rc):
	doc = getClusterStatusBatch(rc)
	if not doc:
		try:
			luci_log.debug_verbose('getClusterStatusBatch returned None for %s/%s' % rc.cluster_info())
		except:
			pass
		return {}

	results = list()

	vals = {}
	vals['type'] = "cluster"

	try:
		vals['alias'] = doc.firstChild.getAttribute('alias')
	except AttributeError, e:
		vals['alias'] = doc.firstChild.getAttribute('name')

	vals['votes'] = doc.firstChild.getAttribute('votes')
	vals['name'] = doc.firstChild.getAttribute('name')
	vals['minQuorum'] = doc.firstChild.getAttribute('minQuorum')
	vals['quorate'] = doc.firstChild.getAttribute('quorate')
	results.append(vals)

	for node in doc.firstChild.childNodes:
		if node.nodeName == 'node':
			vals = {}
			vals['type'] = "node"
			vals['clustered'] = node.getAttribute('clustered')
			vals['name'] = node.getAttribute('name')
			vals['online'] = node.getAttribute('online')
			vals['uptime'] = node.getAttribute('uptime')
			vals['votes'] = node.getAttribute('votes')
			results.append(vals)
		elif node.nodeName == 'service':
			vals = {}
			vals['type'] = 'service'
			vals['name'] = node.getAttribute('name')
			vals['nodename'] = node.getAttribute('nodename')
			vals['running'] = node.getAttribute('running')
			vals['failed'] = node.getAttribute('failed')
			vals['autostart'] = node.getAttribute('autostart')
			results.append(vals)
	return results

def getServicesInfo(self, status, modelb, req):
	map = {}
	maplist = list()

	try:
		baseurl = req['URL']
		if not baseurl:
			raise KeyError, 'is blank'
	except KeyError, e:
		baseurl = '.'

	try:
		cluname = req['clustername']
		if not cluname:
			raise KeyError, 'is blank'
	except KeyError, e:
		try:
			cluname = req.form['clusterName']
			if not cluname:
				raise
		except:
			cluname = '[error retrieving cluster name]'

	for item in status:
		if item['type'] == "service":
			itemmap = {}
			itemmap['name'] = item['name']
			if item['running'] == "true":
				itemmap['running'] = "true"
				itemmap['nodename'] = item['nodename']
			itemmap['autostart'] = item['autostart']
			itemmap['cfgurl'] = baseurl + "?" + "clustername=" + cluname + "&servicename=" + item['name'] + "&pagetype=" + SERVICE

			svc = modelb.retrieveServiceByName(item['name'])
			dom = svc.getAttribute("domain")
			if dom != None:
				itemmap['faildom'] = dom
			else:
				itemmap['faildom'] = "No Failover Domain"
			maplist.append(itemmap)

	map['services'] = maplist
	return map

def getServiceInfo(self, status, modelb, req):
	#set up struct for service config page
	hmap = {}
	root_uuid = 'toplevel'

	try:
		baseurl = req['URL']
		if not baseurl:
			raise KeyError, 'is blank'
	except KeyError, e:
		baseurl = '.'

	try:
		cluname = req['clustername']
		if not cluname:
			raise KeyError, 'is blank'
	except KeyError, e:
		try:
			cluname = req.form['clusterName']
			if not cluname:
				raise
		except:
			cluname = '[error retrieving cluster name]'

	hmap['root_uuid'] = root_uuid
	# uuids for the service page needed when new resources are created
	hmap['uuid_list'] = map(lambda x: make_uuid('resource'), xrange(30))

	try:
		servicename = req['servicename']
	except KeyError, e:
		hmap['resource_list'] = {}
		return hmap

	for item in status:
		if item['type'] == "service":
			if item['name'] == servicename:
				hmap['name'] = servicename
				starturls = list()
				if item['running'] == "true":
					hmap['running'] = "true"
					#In this case, determine where it can run...
					innermap = {}
					nodename = item['nodename']
					innermap['current'] = "This service is currently running on %s" % nodename
					innermap['disableurl'] = baseurl + "?" + "clustername=" + cluname +"&servicename=" + servicename + "&pagetype=" + SERVICE_STOP
					innermap['restarturl'] = baseurl + "?" + "clustername=" + cluname +"&servicename=" + servicename + "&pagetype=" + SERVICE_RESTART
					nodes = modelb.getNodes()
					for node in nodes:
						starturl = {}
						if node.getName() != nodename:
							starturl['nodename'] = node.getName()
							starturl['url'] = baseurl + "?" + "clustername=" + cluname +"&servicename=" + servicename + "&pagetype=" + SERVICE_START + "&nodename=" + node.getName()
							starturls.append(starturl)
					innermap['links'] = starturls
				else:
					#Do not set ['running'] in this case...ZPT will detect it is missing
					#In this case, determine where it can run...
					innermap = {}
					innermap['current'] = "This service is currently stopped"
					innermap['enableurl'] = baseurl + "?" + "clustername=" + cluname +"&servicename=" + servicename + "&pagetype=" + SERVICE_START
					nodes = modelb.getNodes()
					starturls = list()
					for node in nodes:
						starturl = {}
						starturl['nodename'] = node.getName()
						starturl['url'] = baseurl + "?" + "clustername=" + cluname +"&servicename=" + servicename + "&pagetype=" + SERVICE_START + "&nodename=" + node.getName()
						starturls.append(starturl)
					innermap['links'] = starturls
				hmap['innermap'] = innermap

	#Now build hashes for resources under service.
	#first get service by name from model
	svc = modelb.getService(servicename)
	resource_list = list()
	if svc != None:
		indent_ctr = 0
		children = svc.getChildren()
		for child in children:
			recurse_resources(root_uuid, child, resource_list, indent_ctr)

	hmap['resource_list'] = resource_list
	return hmap

def recurse_resources(parent_uuid, child, resource_list, indent_ctr, parent=None):
	#First, add the incoming child as a resource
	#Next, check for children of it
	#Call yourself on every children
	#then return
	rc_map = {}
	if parent != None:
		rc_map['parent'] = parent
	rc_map['name'] = child.getName()
	if child.isRefObject() == True:
		rc_map['ref_object'] = True
		rc_map['type'] = child.getObj().getResourceType()
	else:
		rc_map['type'] = child.getResourceType()

	rc_map['indent_ctr'] = indent_ctr

	#Note: Final version needs all resource attrs
	rc_map['attrs'] = child.getAttributes()
	rc_map['uuid'] = make_uuid('resource')
	rc_map['parent_uuid'] = parent_uuid

	resource_list.append(rc_map)
	kids = child.getChildren()
	child_depth = 0
	new_indent_ctr = indent_ctr + 1
	for kid in kids:
		cdepth = recurse_resources(rc_map['uuid'], kid, resource_list, new_indent_ctr, child)
		child_depth = max(cdepth, child_depth)

	rc_map['max_depth'] = child_depth
	return child_depth + 1

def serviceStart(self, rc, req):
	try:
		svcname = req['servicename']
	except KeyError, e:
		try:
			svcname = req.form['servicename']
		except:
			luci_log.debug_verbose('serviceStart error: no service name')
			return None

	try:
		nodename = req['nodename']
	except KeyError, e:
		try:
			nodename = req.form['nodename']
		except:
			nodename = None

	cluname = None
	try:
		cluname = req['clustername']
	except KeyError, e:
		try:
			cluname = req.form['clustername']
		except:
			pass

	if cluname is None:
		luci_log.debug_verbose('serviceStart error: %s no service name' \
			% svcname)
		return None

	ricci_agent = rc.hostname()

	batch_number, result = startService(rc, svcname, nodename)
	if batch_number is None or result is None:
		luci_log.debug_verbose('startService %s call failed' \
			% svcname)
		return None

	#Now we need to create a DB flag for this system.
	path = str(CLUSTER_FOLDER_PATH + cluname)
	batch_id = str(batch_number)
	objname = str(ricci_agent + "____flag")

	try:
		clusterfolder = self.restrictedTraverse(path)
		clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
		#Now we need to annotate the new DB object
		objpath = str(path + "/" + objname)
		flag = self.restrictedTraverse(objpath)
		flag.manage_addProperty(BATCH_ID, batch_id, "string")
		flag.manage_addProperty(TASKTYPE, SERVICE_START, "string")
		flag.manage_addProperty(FLAG_DESC, "Starting service \'" + svcname + "\'", "string")
	except Exception, e:
		luci_log.debug_verbose('Error creating flag at %s: %s' % (objpath, str(e)))

	response = req.RESPONSE
	response.redirect(req['HTTP_REFERER'] + "&busyfirst=true")

def serviceRestart(self, rc, req):
	try:
		svcname = req['servicename']
	except KeyError, e:
		try:
			svcname = req.form['servicename']
		except:
			luci_log.debug_verbose('no service name for serviceRestart')
			return None
	except:
		luci_log.debug_verbose('no service name for serviceRestart')
		return None

	#Now we need to create a DB flag for this system.
	cluname = None
	try:
		cluname = req['clustername']
	except:
		try:
			cluname = req.form['clustername']
		except:
			pass

	if cluname is None:
		luci_log.debug_verbose('unable to determine cluser name for serviceRestart %s' % svcname)
		return None

	batch_number, result = restartService(rc, svcname)
	if batch_number is None or result is None:
		luci_log.debug_verbose('restartService for %s failed' % svcname)
		return None
				
	ricci_agent = rc.hostname()

	path = str(CLUSTER_FOLDER_PATH + cluname)
	batch_id = str(batch_number)
	objname = str(ricci_agent + "____flag")

	try:
		clusterfolder = self.restrictedTraverse(path)
		clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)

		#Now we need to annotate the new DB object
		objpath = str(path + "/" + objname)
		flag = self.restrictedTraverse(objpath)
		flag.manage_addProperty(BATCH_ID, batch_id, "string")
		flag.manage_addProperty(TASKTYPE, SERVICE_RESTART, "string")
		flag.manage_addProperty(FLAG_DESC, "Restarting service " + svcname, "string")
	except Exception, e:
		luci_log.debug_verbose('Error creating flag in restartService %s: %s' \
			% (svcname, str(e)))

	response = req.RESPONSE
	response.redirect(req['HTTP_REFERER'] + "&busyfirst=true")

def serviceStop(self, rc, req):
	try:
		svcname = req['servicename']
	except KeyError, e:
		try:
			svcname = req.form['servicename']
		except:
			luci_log.debug_verbose('no service name for serviceStop')
			return None
	except:
		luci_log.debug_verbose('no service name for serviceStop')
		return None

	#Now we need to create a DB flag for this system.
	cluname = None
	try:
		cluname = req['clustername']
	except:
		try:
			cluname = req.form['clustername']
		except:
			pass

	if cluname is None:
		luci_log.debug_verbose('unable to determine cluser name for serviceStop %s' % svcname)
		return None

	batch_number, result = stopService(rc, svcname)
	if batch_number is None or result is None:
		luci_log.debug_verbose('stopService for %s failed' % svcname)
		return None

	ricci_agent = rc.hostname()

	path = str(CLUSTER_FOLDER_PATH + cluname)
	batch_id = str(batch_number)
	objname = str(ricci_agent + "____flag")

	try:
		clusterfolder = self.restrictedTraverse(path)
		clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
		#Now we need to annotate the new DB object
		objpath = str(path + "/" + objname)
		flag = self.restrictedTraverse(objpath)

		flag.manage_addProperty(BATCH_ID, batch_id, "string")
		flag.manage_addProperty(TASKTYPE, SERVICE_STOP, "string")
		flag.manage_addProperty(FLAG_DESC, "Stopping service " + svcname, "string")
		time.sleep(2)
	except Exception, e:
		luci_log.debug_verbose('Error creating flags for stopService %s: %s' \
			% (svcname, str(e)))

	response = req.RESPONSE
	response.redirect(req['HTTP_REFERER'] + "&busyfirst=true")

def getFdomsInfo(self, modelb, request, clustatus):
  slist = list()
  nlist = list()
  for item in clustatus:
    if item['type'] == "node":
      nlist.append(item)
    elif item['type'] == "service":
      slist.append(item)
  fdomlist = list()
  clustername = request['clustername']
  baseurl = request['URL']
  fdoms = modelb.getFailoverDomains()
  svcs = modelb.getServices()
  for fdom in fdoms:
    fdom_map = {}
    fdom_map['name'] = fdom.getName()
    fdom_map['cfgurl'] = baseurl + "?pagetype=" + FDOM_LIST + "&clustername=" + clustername
    ordered_attr = fdom.getAttribute('ordered')
    restricted_attr = fdom.getAttribute('restricted')
    if ordered_attr != None and (ordered_attr == "true" or ordered_attr == "1"):
      fdom_map['ordered'] = True
    else:
      fdom_map['ordered'] = False
    if restricted_attr != None and (restricted_attr == "true" or restricted_attr == "1"):
      fdom_map['restricted'] = True
    else:
      fdom_map['restricted'] = False
    nodes = fdom.getChildren()
    nodelist = list()
    for node in nodes:
      nodesmap = {}
      ndname = node.getName()
      for nitem in nlist:
        if nitem['name'] == ndname:
          break
      nodesmap['nodename'] = ndname
      nodesmap['nodecfgurl'] = baseurl + "?clustername=" + clustername + "&nodename=" + ndname + "&pagetype=" + NODE
      if nitem['clustered'] == "true":
        nodesmap['status'] = NODE_ACTIVE
      elif nitem['online'] == "false":
        nodesmap['status'] = NODE_UNKNOWN
      else:
        nodesmap['status'] = NODE_INACTIVE
      priority_attr =  node.getAttribute('priority')
      if priority_attr != None:
        nodesmap['priority'] = "0"
      nodelist.append(nodesmap)
    fdom_map['nodeslist'] = nodelist

    svclist = list()
    for svc in svcs:
      svcname = svc.getName()
      for sitem in slist:
        if sitem['name'] == svcname:
          break  #found more info about service...

      domain = svc.getAttribute("domain")
      if domain != None:
        if domain == fdom.getName():
          svcmap = {}
          svcmap['name'] = svcname
          svcmap['status'] = sitem['running']
          svcmap['svcurl'] = baseurl + "?pagetype=" + SERVICE + "&clustername=" + clustername + "&servicename=" + svcname
          svcmap['location'] = sitem['nodename']
          svclist.append(svcmap)
    fdom_map['svclist'] = svclist
    fdomlist.append(fdom_map)
  return fdomlist

def processClusterProps(self, ricci_agent, request):
  #First, retrieve cluster.conf from session
  conf = request.SESSION.get('conf')
  model = ModelBuilder(0, None, None, conf)

  #Next, determine actiontype and switch on it
  actiontype = request[ACTIONTYPE]

  if actiontype == BASECLUSTER:
    cp = model.getClusterPtr()
    cfgver = cp.getConfigVersion()

    rcfgver = request['cfgver']

    if cfgver != rcfgver:
      cint = int(cfgver)
      rint = int(rcfgver)
      if rint > cint:
        cp.setConfigVersion(rcfgver)

    rname = request['cluname']
    name = model.getClusterAlias()

    if rname != name:
      cp.addAttribute('alias', rname)

    response = request.RESPONSE
    response.redirect(request['HTTP_REFERER'] + "&busyfirst=true")
    return

  elif actiontype == FENCEDAEMON:
    pass

  elif actiontype == MULTICAST:
    pass

  elif actiontype == QUORUMD:
    pass

  else:
    return


def getClusterInfo(self, model, req):
  cluname = req[CLUNAME]
  baseurl = req['URL'] + "?" + PAGETYPE + "=" + CLUSTER_PROCESS + "&" + CLUNAME + "=" + cluname + "&"
  map = {}
  basecluster_url = baseurl + ACTIONTYPE + "=" + BASECLUSTER
  #needed:
  map['basecluster_url'] = basecluster_url
  #name field
  map['clustername'] = model.getClusterAlias()
  #config version
  cp = model.getClusterPtr()
  map['config_version'] = cp.getConfigVersion()
  #-------------
  #new cluster params - if rhel5
  #-------------
  #Fence Daemon Props
  fencedaemon_url = baseurl + ACTIONTYPE + "=" + FENCEDAEMON
  map['fencedaemon_url'] = fencedaemon_url
  fdp = model.getFenceDaemonPtr()
  pjd = fdp.getAttribute('post_join_delay')
  if pjd == None:
    pjd = "6"
  pfd = fdp.getAttribute('post_fail_delay')
  if pfd == None:
    pfd = "0"
  #post join delay
  map['pjd'] = pjd
  #post fail delay
  map['pfd'] = pfd
  #-------------
  #if multicast
  multicast_url = baseurl + ACTIONTYPE + "=" + MULTICAST
  map['multicast_url'] = multicast_url
  #mcast addr
  is_mcast = model.isMulticast()
  #map['is_mcast'] = is_mcast
  if is_mcast:
    map['mcast_addr'] = model.getMcastAddr()
    map['is_mcast'] = "True"
  else:
    map['is_mcast'] = "False"
    map['mcast_addr'] = "1.2.3.4"

  #-------------
  #quorum disk params
  quorumd_url = baseurl + ACTIONTYPE + "=" + QUORUMD
  map['quorumd_url'] = quorumd_url
  is_quorumd = model.isQuorumd()
  map['is_quorumd'] = is_quorumd
  map['interval'] = ""
  map['tko'] = ""
  map['votes'] = ""
  map['min_score'] = ""
  map['device'] = ""
  map['label'] = ""

  #list struct for heuristics...
  hlist = list()

  if is_quorumd:
    qdp = model.getQuorumdPtr()
    interval = qdp.getAttribute('interval')
    if interval != None:
      map['interval'] = interval

    tko = qdp.getAttribute('tko')
    if tko != None:
      map['tko'] = tko

    votes = qdp.getAttribute('votes')
    if votes != None:
      map['votes'] = votes

    min_score = qdp.getAttribute('min_score')
    if min_score != None:
      map['min_score'] = min_score

    device = qdp.getAttribute('device')
    if device != None:
      map['device'] = device

    label = qdp.getAttribute('label')
    if label != None:
      map['label'] = label

    heuristic_kids = qdp.getChildren()
    h_ctr = 0
    for kid in heuristic_kids:
      hmap = {}
      hname = kid.getAttribute('name')
      if hname == None:
        hname = h_ctr
        h_ctr = h_ctr + 1
      hprog = kid.getAttribute('program')
      hscore = kid.getAttribute('score')
      hinterval = kid.getAttribute('interval')
      if hprog == None:
        continue
      if hname != None:
        hmap['hname'] = hname
      else:
        hmap['hname'] = ""
      hmap['hprog'] = hprog
      if hscore != None:
        hmap['hscore'] = hscore
      else:
        hmap['hscore'] = ""
      if hinterval != None:
        hmap['hinterval'] = hinterval
      else:
        hmap['hinterval'] = ""
      hlist.append(hmap)
  map['hlist'] = hlist

  return map

def getClustersInfo(self,status,req):
  map = {}
  nodelist = list()
  svclist = list()
  clulist = list()
  baseurl = req['URL']
  for item in status:
    if item['type'] == "node":
      nodelist.append(item)
    elif item['type'] == "service":
      svclist.append(item)
    elif item['type'] == "cluster":
      clulist.append(item)
    else:
      continue
  if len(clulist) < 1:
    return {}
  clu = clulist[0]
  clustername = clu['name']
  if clu['alias'] != "":
    map['clusteralias'] = clu['alias']
  else:
    map['clusteralias'] = clustername
  map['clustername'] = clustername
  if clu['quorate'] == "true":
    map['status'] = "Quorate"
    map['running'] = "true"
  else:
    map['status'] = "Not Quorate"
    map['running'] = "false"
  map['votes'] = clu['votes']
  map['minquorum'] = clu['minQuorum']
  map['clucfg'] = baseurl + "?" + PAGETYPE + "=" + CLUSTER_CONFIG + "&" + CLUNAME + "=" + clustername
  svc_dict_list = list()
  for svc in svclist:
      svc_dict = {}
      svc_dict['nodename'] = svc['nodename']
      svcname = svc['name']
      svc_dict['name'] = svcname
      svc_dict['srunning'] = svc['running']
      svcurl = baseurl + "?" + PAGETYPE + "=" + SERVICE + "&" + CLUNAME + "=" + clustername + "&servicename=" + svcname
      svc_dict['servicename'] = svcname
      svc_dict['svcurl'] = svcurl
      svc_dict_list.append(svc_dict)
  map['currentservices'] = svc_dict_list
  node_dict_list = list()
  for item in nodelist:
    nmap = {}
    name = item['name']
    nmap['nodename'] = name
    cfgurl = baseurl + "?" + PAGETYPE + "=" + NODE + "&" + CLUNAME + "=" + clustername + "&nodename=" + name
    nmap['configurl'] = cfgurl
    if item['clustered'] == "true":
      nmap['status'] = NODE_ACTIVE
    elif item['online'] == "false":
      nmap['status'] = NODE_UNKNOWN
    else:
      nmap['status'] = NODE_INACTIVE
    node_dict_list.append(nmap)

  map['currentnodes'] = node_dict_list

  return map

def nodeTaskProcess(self, model, request):
	try:
		clustername = request['clustername']
	except KeyError, e:
		try:
			clustername = request.form['clustername']
		except:
			luci_log.debug('missing cluster name for NTP')
			return None

	try:
		nodename = request['nodename']
	except KeyError, e:
		try:
			nodename = request.form['nodename']
		except:
			luci_log.debug('missing nodename name for NTP')
			return None

	try:
		task = request['task']
	except KeyError, e:
		try:
			task = request.form['task']
		except:
			luci_log.debug('missing task for NTP')
			return None

	nodename_resolved = resolve_nodename(self, clustername, nodename)
	if not nodename_resolved or not nodename or not task or not clustername:
		luci_log.debug('resolve_nodename failed for NTP')
		return None

	if task != NODE_FENCE:
		# Fencing is the only task for which we don't
		# want to talk to the node on which the action is
		# to be performed.
		try:
			rc = RicciCommunicator(nodename_resolved)
		except RicciError, e:
			luci_log.debug('ricci error from %s: %s' \
				% (nodename_resolved, str(e)))
			return None
		except:
			return None

		cluinfo = rc.cluster_info()
		if not cluinfo[0] and not cluinfo[1]:
			luci_log.debug('host %s not in a cluster (expected %s)' \
				% (nodename_resolved, clustername))
			return None

		cname = lower(clustername)
		if cname != lower(cluinfo[0]) and cname != lower(cluinfo[1]):
			luci_log.debug('host %s in unknown cluster %s:%s (expected %s)' \
				% (nodename_resolved, cluinfo[0], cluinfo[1], clustername))
			return None

		if not rc.authed():
			rc = None
			try:
				snode = getStorageNode(self, nodename)
				setNodeFlag(snode, CLUSTER_NODE_NEED_AUTH)
			except:
				# we'll hit it again, and try again then
				pass

			try:
				cnode = getClusterNode(self, nodename, clustername)
				setNodeFlag(cnode, CLUSTER_NODE_NEED_AUTH)
			except:
				# we'll hit it again, and try again then
				pass

		if rc is None:
			return None

	if task == NODE_LEAVE_CLUSTER:
		path = str(CLUSTER_FOLDER_PATH + clustername + "/" + nodename_resolved)

		try:
			nodefolder = self.restrictedTraverse(path)
			if not nodefolder:
				raise Exception, 'cannot find directory at %s' % path
		except Exception, e:
			luci_log.debug('node_leave_cluster err: %s' % str(e))
			return None

		objname = str(nodename_resolved + "____flag")

		fnpresent = noNodeFlagsPresent(self, nodefolder, objname, nodename_resolved)
		if fnpresent is None:
			luci_log.debug('An error occurred while checking flags for %s' \
				% nodename_resolved)
			return None

		if fnpresent == False:
			luci_log.debug('flags are still present for %s -- bailing out' \
				% nodename_resolved)
			return None

		batch_number, result = nodeLeaveCluster(rc)
		if batch_number is None or result is None:
			luci_log.debug_verbose('nodeLeaveCluster error: batch_number and/or result is None')
			return None

		batch_id = str(batch_number)
		objpath = str(path + "/" + objname)

		try:
			nodefolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
			#Now we need to annotate the new DB object
			flag = self.restrictedTraverse(objpath)
			flag.manage_addProperty(BATCH_ID, batch_id, "string")
			flag.manage_addProperty(TASKTYPE, NODE_LEAVE_CLUSTER, "string")
			flag.manage_addProperty(FLAG_DESC, "Node \'" + nodename + "\' leaving cluster", "string")
		except:
			luci_log.debug('An error occurred while setting flag %s' % objpath)

		response = request.RESPONSE
		#Is this correct? Should we re-direct to the cluster page?
		response.redirect(request['URL'] + "?pagetype=" + CLUSTER_CONFIG + "&clustername=" + clustername)
	elif task == NODE_JOIN_CLUSTER:
		batch_number, result = nodeJoinCluster(rc)
		if batch_number is None or result is None:
			luci_log.debug_verbose('nodeJoin error: batch_number and/or result is None')
			return None

		path = str(CLUSTER_FOLDER_PATH + clustername + "/" + nodename_resolved)
		batch_id = str(batch_number)
		objname = str(nodename_resolved + "____flag")
		objpath = str(path + "/" + objname)

		try:
			nodefolder = self.restrictedTraverse(path)
			nodefolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
			#Now we need to annotate the new DB object
			flag = self.restrictedTraverse(objpath)
			flag.manage_addProperty(BATCH_ID, batch_id, "string")
			flag.manage_addProperty(TASKTYPE, NODE_JOIN_CLUSTER, "string")
			flag.manage_addProperty(FLAG_DESC, "Node \'" + nodename + "\' joining cluster", "string")
		except Exception, e:
			luci_log.debug_verbose('nodeJoin error: creating flags at %s: %s' \
				% (path, str(e)))

		response = request.RESPONSE
		#Once again, is this correct? Should we re-direct to the cluster page?
		response.redirect(request['URL'] + "?pagetype=" + CLUSTER_CONFIG + "&clustername=" + clustername)
	elif task == NODE_REBOOT:
		batch_number, result = nodeReboot(rc)
		if batch_number is None or result is None:
			luci_log.debug_verbose('nodeReboot: batch_number and/or result is None')
			return None

		path = str(CLUSTER_FOLDER_PATH + clustername + "/" + nodename_resolved)
		batch_id = str(batch_number)
		objname = str(nodename_resolved + "____flag")
		objpath = str(path + "/" + objname)

		try:
			nodefolder = self.restrictedTraverse(path)
			nodefolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
			#Now we need to annotate the new DB object
			flag = self.restrictedTraverse(objpath)
			flag.manage_addProperty(BATCH_ID, batch_id, "string")
			flag.manage_addProperty(TASKTYPE, NODE_REBOOT, "string")
			flag.manage_addProperty(FLAG_DESC, "Node \'" + nodename + "\' is being rebooted", "string")
		except Exception, e:
			luci_log.debug_verbose('nodeReboot err: creating flags at %s: %s' \
				% (path, str(e)))

		response = request.RESPONSE
		#Once again, is this correct? Should we re-direct to the cluster page?
		response.redirect(request['URL'] + "?pagetype=" + CLUSTER_CONFIG + "&clustername=" + clustername)
	elif task == NODE_FENCE:
		#here, we DON'T want to open connection to node to be fenced.
		path = str(CLUSTER_FOLDER_PATH + clustername)
		try:
			clusterfolder = self.restrictedTraverse(path)
			if not clusterfolder:
				raise Exception, 'no cluster folder at %s' % path
		except Exception, e:
			luci_log.debug('The cluster folder for %s could not be found: %s' \
				 % (clustername, str(e)))
			return None

		try:
			nodes = clusterfolder.objectItems('Folder')
			if not nodes or len(nodes) < 1:
				raise Exception, 'no cluster nodes'
		except Exception, e:
			luci_log.debug('No cluster nodes for %s were found: %s' \
				% (clustername, str(e)))
			return None

		found_one = False
		for node in nodes:
			if node[1].getId().find(nodename) != (-1):
				continue

			try:
				rc = RicciCommunicator(node[1].getId())
				if not rc:
					continue
			except RicciError, e:
				luci_log.debug('ricci error for host %s: %s' \
					% (node[0], str(e)))
				continue
			except:
				continue

			if not rc.authed():
				rc = None
				try:
					snode = getStorageNode(self, node[1].getId())
					setNodeFlag(snode, CLUSTER_NODE_NEED_AUTH)
				except:
					pass

				try:
					setNodeFlag(node[1], CLUSTER_NODE_NEED_AUTH)
				except:
					pass

				continue
			found_one = True
			break

		if not found_one:
			return None

		batch_number, result = nodeFence(rc, nodename)
		if batch_number is None or result is None:
			luci_log.debug_verbose('nodeFence: batch_number and/or result is None')
			return None

		path = str(path + "/" + nodename_resolved)
		batch_id = str(batch_number)
		objname = str(nodename_resolved + "____flag")
		objpath = str(path + "/" + objname)

		try:
			nodefolder = self.restrictedTraverse(path)
			nodefolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
			#Now we need to annotate the new DB object
			flag = self.restrictedTraverse(objpath)
			flag.manage_addProperty(BATCH_ID, batch_id, "string")
			flag.manage_addProperty(TASKTYPE, NODE_FENCE, "string")
			flag.manage_addProperty(FLAG_DESC, "Node \'" + nodename + "\' is being fenced", "string")
		except Exception, e:
			luci_log.debug_verbose('nodeFence err: creating flags at %s: %s' \
				% (path, str(e)))

		response = request.RESPONSE
		#Once again, is this correct? Should we re-direct to the cluster page?
		response.redirect(request['URL'] + "?pagetype=" + CLUSTER_CONFIG + "&clustername=" + clustername)
	elif task == NODE_DELETE:
		#We need to get a node name other than the node
		#to be deleted, then delete the node from the cluster.conf
		#and propogate it. We will need two ricci agents for this task.

		# Make sure we can find a second node before we hose anything.
		path = str(CLUSTER_FOLDER_PATH + clustername)
		try:
			clusterfolder = self.restrictedTraverse(path)
			if not clusterfolder:
				raise Exception, 'no cluster folder at %s' % path
		except Exception, e:
			luci_log.debug_verbose('node delete error for cluster %s: %s' \
				% (clustername, str(e)))
			return None

		try:
			nodes = clusterfolder.objectItems('Folder')
			if not nodes or len(nodes) < 1:
				raise Exception, 'no cluster nodes in DB'
		except Exception, e:
			luci_log.debug_verbose('node delete error for cluster %s: %s' \
				% (clustername, str(e)))

		found_one = False
		for node in nodes:
			if node[1].getId().find(nodename) != (-1):
				continue
			#here we make certain the node is up...
			# XXX- we should also make certain this host is still
			# in the cluster we believe it is.
			try:
				rc2 = RicciCommunicator(node[1].getId())
			except Exception, e:
				luci_log.info('ricci %s error: %s' % (node[0], str(e)))
				continue
			except:
				continue

			if not rc2.authed():
				try:
					setNodeFlag(node[1], CLUSTER_NODE_NEED_AUTH)
				except:
					pass

				try:
					snode = getStorageNode(self, node[0])
					setNodeFlag(snode, CLUSTER_NODE_NEED_AUTH)
				except:
					pass

				luci_log.debug_verbose('%s is not authed' % node[0])
				rc2 = None
				continue
			else:
				found_one = True
				break

		if not found_one:
			luci_log.debug_verbose('unable to find ricci node to delete %s from %s' % (nodename, clustername))
			return None

		#First, delete cluster.conf from node to be deleted.
		#next, have node leave cluster.
		batch_number, result = nodeLeaveCluster(rc, purge=True)
		if batch_number is None or result is None:
			luci_log.debug_verbose('nodeDelete: batch_number and/or result is None')
			return None

		#It is not worth flagging this node in DB, as we are going
		#to delete it anyway. Now, we need to delete node from model
		#and send out new cluster.conf
		delete_target = None
		nodelist = model.getNodes()
		find_node = lower(nodename)
		for n in nodelist:
			try:
				if lower(n.getName()) == find_node:
					delete_target = n
					break
			except:
				continue

		if delete_target is None:
			luci_log.debug_verbose('unable to find delete target for %s in %s' \
				% (nodename, clustername))
			return None

		model.deleteNode(delete_target)

		try:
			str_buf = model.exportModelAsString()
			if not str_buf:
				raise Exception, 'model string is blank'
		except Exception, e:
			luci_log.debug_verbose('NTP exportModelAsString: %s' % str(e))
			return None

		# propagate the new cluster.conf via the second node
		batch_number, result = setClusterConf(rc2, str(str_buf))
		if batch_number is None:
			luci_log.debug_verbose('batch number is None after del node in NTP')
			return None

		#Now we need to delete the node from the DB
		path = str(CLUSTER_FOLDER_PATH + clustername)
		del_path = str(path + "/" + nodename_resolved)

		try:
			delnode = self.restrictedTraverse(del_path)
			clusterfolder = self.restrictedTraverse(path)
			clusterfolder.manage_delObjects(delnode[0])
		except Exception, e:
			luci_log.debug_verbose('error deleting %s: %s' % (del_path, str(e)))

		batch_id = str(batch_number)
		objname = str(nodename_resolved + "____flag")
		objpath = str(path + "/" + objname)

		try:
			clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
			#Now we need to annotate the new DB object
			flag = self.restrictedTraverse(objpath)
			flag.manage_addProperty(BATCH_ID, batch_id, "string")
			flag.manage_addProperty(TASKTYPE, NODE_DELETE, "string")
			flag.manage_addProperty(FLAG_DESC, "Deleting node \'" + nodename + "\'", "string")
		except Exception, e:
			luci_log.debug_verbose('nodeDelete %s err setting flag at %s: %s' \
				% (nodename, objpath, str(e)))

		response = request.RESPONSE
		response.redirect(request['HTTP_REFERER'] + "&busyfirst=true")

def getNodeInfo(self, model, status, request):
  infohash = {}
  baseurl = request['URL']
  nodestate = NODE_ACTIVE
  svclist = list()
  for thing in status:
    if thing['type'] == "service":
      svclist.append(thing)

  #Get cluster name and node name from request
  clustername = request['clustername']
  nodename = request['nodename']
  #extract correct node line from cluster status
  found = False
  for item in status:
    if (item['type'] == "node") and (item['name'] == nodename):
      found = True
      break
  if found == False:
    raise UnknownClusterError("Fatal", "Unable to resolve node name in cluster status")
  #Now determine state of node...
  if item['online'] == "false":
    nodestate = NODE_UNKNOWN
  elif item['clustered'] == "true":
    nodestate = NODE_ACTIVE
  else:
    nodestate = NODE_INACTIVE

  infohash['nodestate'] = nodestate
  infohash['nodename'] = nodename

  #set up drop down links
  if nodestate == NODE_ACTIVE:
    infohash['jl_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_LEAVE_CLUSTER + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['reboot_url'] = baseurl + "?pagetype=" +NODE_PROCESS + "&task=" + NODE_REBOOT + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['fence_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_FENCE + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['delete_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_DELETE + "&nodename=" + nodename + "&clustername=" + clustername

  if nodestate == NODE_INACTIVE:
    infohash['jl_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_JOIN_CLUSTER + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['reboot_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_REBOOT + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['fence_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_FENCE + "&nodename=" + nodename + "&clustername=" + clustername
    infohash['delete_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_DELETE + "&nodename=" + nodename + "&clustername=" + clustername

  #figure out current services running on this node
  svc_dict_list = list()
  for svc in svclist:
    if svc['nodename'] == nodename:
      svc_dict = {}
      svcname = svc['name']
      svcurl = baseurl + "?" + PAGETYPE + "=" + SERVICE + "&" + CLUNAME + "=" + clustername + "&servicename=" + svcname
      svc_dict['servicename'] = svcname
      svc_dict['svcurl'] = svcurl
      svc_dict_list.append(svc_dict)

  infohash['currentservices'] = svc_dict_list

  #next is faildoms
  fdoms = model.getFailoverDomainsForNode(nodename)
  fdom_dict_list = list()
  for fdom in fdoms:
    fdom_dict = {}
    fdom_dict['name'] = fdom.getName()
    fdomurl = baseurl + "?" + PAGETYPE + "=" + FDOM_CONFIG + "&" + CLUNAME + "=" + clustername + "&fdomname=" + fdom.getName()
    fdom_dict['fdomurl'] = fdomurl
    fdom_dict_list.append(fdom_dict)

  infohash['fdoms'] = fdom_dict_list

  #return infohash
  infohash['d_states'] = None
  if nodestate == NODE_ACTIVE or nodestate == NODE_INACTIVE:
  #call service module on node and find out which daemons are running
    rc = RicciCommunicator(nodename)
    dlist = list()
    dlist.append("ccsd")
    dlist.append("cman")
    dlist.append("fenced")
    dlist.append("rgmanager")
    states = getDaemonStates(rc, dlist)
    infohash['d_states'] = states

  infohash['logurl'] = '/luci/logs/?nodename=' + nodename + '&clustername=' + clustername
  return infohash
  #get list of faildoms for node

def getNodesInfo(self, model,status,req):
  resultlist = list()
  nodelist = list()
  svclist = list()
  #Sort into lists...
  for item in status:
    if item['type'] == "node":
      nodelist.append(item)
    elif item['type'] == "service":
      svclist.append(item)
    else:
      continue

  for item in nodelist:
    map = {}
    name = item['name']
    map['nodename'] = name
    clustername = req['clustername']
    baseurl = req['URL']
    cfgurl = baseurl + "?" + PAGETYPE + "=" + NODE + "&" + CLUNAME + "=" + clustername + "&nodename=" + name
    map['configurl'] = cfgurl
    map['fenceurl'] = cfgurl + "#fence"
    if item['clustered'] == "true":
      map['status'] = NODE_ACTIVE
      map['status_str'] = NODE_ACTIVE_STR
    elif item['online'] == "false":
      map['status'] = NODE_UNKNOWN
      map['status_str'] = NODE_UNKNOWN_STR
    else:
      map['status'] = NODE_INACTIVE
      map['status_str'] = NODE_INACTIVE_STR

    map['logurl'] = '/luci/logs?nodename=' + name + '&clustername=' + clustername
    #set up URLs for dropdown menu...
    if map['status'] == NODE_ACTIVE:
      map['jl_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_LEAVE_CLUSTER + "&nodename=" + name + "&clustername=" + clustername
      map['reboot_url'] = baseurl + "?pagetype=" +NODE_PROCESS + "&task=" + NODE_REBOOT + "&nodename=" + name + "&clustername=" + clustername
      map['fence_it_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_FENCE + "&nodename=" + name + "&clustername=" + clustername
      map['delete_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_DELETE + "&nodename=" + name + "&clustername=" + clustername

    if map['status'] == NODE_INACTIVE:
      map['jl_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_JOIN_CLUSTER + "&nodename=" + name + "&clustername=" + clustername
      map['reboot_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_REBOOT + "&nodename=" + name + "&clustername=" + clustername
      map['fence_it_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_FENCE + "&nodename=" + name + "&clustername=" + clustername
      map['delete_url'] = baseurl + "?pagetype=" + NODE_PROCESS + "&task=" + NODE_DELETE + "&nodename=" + name + "&clustername=" + clustername


    #figure out current services running on this node
    svc_dict_list = list()
    for svc in svclist:
      if svc['nodename'] == name:
        svc_dict = {}
        svcname = svc['name']
        svcurl = baseurl + "?" + PAGETYPE + "=" + SERVICE + "&" + CLUNAME + "=" + clustername + "&servicename=" + svcname
        svc_dict['servicename'] = svcname
        svc_dict['svcurl'] = svcurl
        svc_dict_list.append(svc_dict)

    map['currentservices'] = svc_dict_list
    #next is faildoms
    fdoms = model.getFailoverDomainsForNode(name)
    fdom_dict_list = list()
    for fdom in fdoms:
      fdom_dict = {}
      fdom_dict['name'] = fdom.getName()
      fdomurl = baseurl + "?" + PAGETYPE + "=" + FDOM_CONFIG + "&" + CLUNAME + "=" + clustername + "&fdomname=" + fdom.getName()
      fdom_dict['fdomurl'] = fdomurl
      fdom_dict_list.append(fdom_dict)

    map['fdoms'] = fdom_dict_list
    resultlist.append(map)

  return resultlist

def getFence(self, model, request):
  map = {}
  fencename = request['fencedevicename']
  fencedevs = model.getFenceDevices()
  for fencedev in fencedevs:
    if fencedev.getName().strip() == fencename:
      map = fencedev.getAttributes()
      return map

  return map
  
def getFenceInfo(self, model, request):
  map = {}
  fencedevs = list() 
  level1 = list()
  level2 = list()
  map['level1'] = level1
  map['level2'] = level2
  map['fencedevs'] = fencedevs
  nodename = ""
  if request == None:  #this is being called by the fence device page
    #Get list of fence devices
    fds = model.getFenceDevices()
    for fd in fds:
      #create fencedev hashmap
      if fd.isShared() == True:
        fencedev = fd.getAttributes()
        fencedevs.append(fencedev)
      
    return map

  else:
    try:
      nodename = request['nodename']
    except KeyError, e:
      raise GeneralError('FATAL', "Could not extract nodename from request")
    
    #here we need to get fences for a node - just the first two levels
    #then fill in two data structures with all attr's 
    try:
      node = model.retrieveNodeByName(nodename)
    except GeneralError, e:
      raise GeneralError('FATAL', "Couldn't find node name in current node list")

    levels = node.getFenceLevels()
    len_levels = len(levels)

    if len_levels == 0:
      return map

    for i in xrange(2):
      if not i in levels:
        continue
      fence_struct = {}
      if levels[i] != None:
        level = levels[i]
      else:
        #No more levels...
        continue
      kids = level.getChildren()
      if len(kids) == 0:
        continue
      else:
        #for each kid, 
        ### resolve name, find fence device
        ### Add fd to list, if it is not there yet 
        ### determine if it is a shared fence type
        ### if it is a shared device, add instance entry
        fds = model.getFenceDevices()
        fence_struct = None
        for kid in kids:
          name = kid.getName()
          found_fd = False
          if not i in map:
			continue
          for entry in map[i]:
            if entry['name'] == name:
              fence_struct = entry
              found_fd = True
              break
          if found_fd == False:
            for fd in fds:
              if fd.getName() == name:  #Found the fence device
                fence_struct = {}
                fence_struct['isShareable'] = fd.isShared()
                fd_attrs = fd.getAttributes()
                kees = fd_attrs.keys()
                for kee in kees:
                  fence_struct[kee] = fd_attrs[kee]
          fi_attrs = kid.getAttributes()
          kees = fi_attrs.keys()
          if fence_struct['isShareable'] == True:
            instance_struct = {}
            for kee in kees:
              instance_struct[kee] = fi_attrs[kee]
              try:
                  check = fence_struct['instances']
                  check.append(instance_struct)
              except KeyError, e:
                  fence_struct['instances'] = list()
                  fence_struct['instances'].append(instance_struct) 
          else:  #Not a shareable fence device type
            for kee in kees:
              fence_struct[kee] = fi_attrs[kee]
        if i == 0:
          level1.append(fence_struct)      
        else:
          level2.append(fence_struct)      

    return map    
      
def getLogsForNode(self, request):
	try:
		nodename = request['nodename']
	except KeyError, e:
		try:
			nodename = request.form['nodename']
		except:
			luci_log.debug_verbose('Unable to get node name to retrieve logging information')
			return 'Unable to get node name to retrieve logging information'

	clustername = None
	try:
		clustername = request['clustername']
	except KeyError, e:
		try:
			clustername = request.form['clusterName']
			if not clustername:
				raise
		except:
			clustername = None
			luci_log.debug_verbose('Unable to find cluster name while retrieving logging information for %s' % nodename)
	except:
		pass

	if clustername is None:
		nodename_resolved = nodename
	else:
		try:
			nodename_resolved = resolve_nodename(self, clustername, nodename)
		except:
			luci_log.debug_verbose('Unable to resolve node name %s/%s to retrieve logging information' \
				% (nodename, clustername))
			return 'Unable to resolve node name for %s in cluster %s' % (nodename, clustername)

	try:
		rc = RicciCommunicator(nodename_resolved)
	except RicciError, e:
		luci_log.debug_verbose('Ricci error while getting logs for %s: %s' \
			% (nodename_resolved, str(e)))
		return 'Ricci error while getting logs for %s' % nodename_resolved
	except:
		luci_log.debug_verbose('Unexpected exception while getting logs for %s' \
			% nodename_resolved)
		return 'Ricci error while getting logs for %s' % nodename_resolved

	if not rc.authed():
		try:
			snode = getStorageNode(self, nodename)
			setNodeFlag(snode, CLUSTER_NODE_NEED_AUTH)
		except:
			pass

		if clustername:
			try:
				cnode = getClusterNode(self, nodename, clustername)
				setNodeFlag(cnode, CLUSTER_NODE_NEED_AUTH)
			except:
				pass

		return 'Luci is not authenticated to node %s. Please reauthenticate first.' % nodename

	return getNodeLogs(rc)

def processXenVM(self, req):
  model = req.SESSION.get('model')
  isNew = False
  try:
    xenvmname = req['servicename']
  except KeyError, e:
    isNew = True
  
  if isNew == True:
    xvm = Vm()
    xvm.addAttribute("name", req.form['xenvmname'])
    xvm.addAttribute("path", req.form['xenvmpath'])
    rmptr = model.getResourceManagerPtr()
    rmptr.addChild(xvm)
  else:
    xvm = model.retrieveXenVMsByName(self, xenvmname)
    xvm.addAttribute("name", req.form['xenvmname'])
    xvm.addAttribute("path", req.form['xenvmpath'])

  try:
    stringbuf = model.exportModelAsString()
    if not stringbuf:
   	  raise Exception, 'model is blank'
  except Exception, e:
    luci_log.debug_verbose('exportModelAsString error: %s' % str(e))
    return None

  try:
    clustername = model.getClusterName()
    if not clustername:
      raise Exception, 'cluster name from modelb.getClusterName() is blank'
  except Exception, e:
    luci_log.debug_verbose('error: getClusterName: %s' % str(e))
    return None

  rc = getRicciAgent(self, clustername)
  if not rc:
    luci_log.debug_verbose('Unable to find a ricci agent for the %s cluster' % clustername)
    return None

  setClusterConf(rc, stringbuf)

def getXenVMInfo(self, model, request):
	try:
		xenvmname = request['servicename']
	except KeyError, e:
		try:
			xenvmname = request.form['servicename']
		except:
			luci_log.debug_verbose('servicename is missing from request')
			return {}
	except:
		luci_log.debug_verbose('servicename is missing from request')
		return {}

	try:  
		xenvm = model.retrieveXenVMsByName(xenvmname)
	except:
		luci_log.debug('An error occurred while attempting to get VM %s' \
			% xenvmname)
		return {}

	map = xenvm.getAttributes()
	return map

def isClusterBusy(self, req):
  items = None
  map = {}
  isBusy = False
  redirect_message = False
  nodereports = list()
  map['nodereports'] = nodereports

  try:
    cluname = req['clustername']
  except KeyError, e:
    try:
      cluname = req.form['clustername']
    except:
      try:
        cluname = req.form['clusterName']
      except:
        luci_log.debug_verbose('ICB0: No cluster name -- returning empty map')
        return map

  path = str(CLUSTER_FOLDER_PATH + cluname)
  try:
    clusterfolder = self.restrictedTraverse(path)
    if not clusterfolder:
      raise Exception, 'clusterfolder is None'
  except Exception, e:
    luci_log.debug_verbose('ICB1: cluster %s [%s] folder missing: %s -- returning empty map' % (cluname, path, str(e)))
    return map
  except:
    luci_log.debug_verbose('ICB2: cluster %s [%s] folder missing: returning empty map' % (cluname, path))

  try:
    items = clusterfolder.objectItems('ManagedSystem')
    if not items or len(items) < 1:
      luci_log.debug_verbose('ICB3: no flags at %s for cluster %s' \
          % (cluname, path))
      return map  #This returns an empty map, and should indicate not busy
  except Exception, e:
    luci_log.debug('ICB4: An error occurred while looking for cluster %s flags at path %s: %s' % (cluname, path, str(e)))
    return map
  except:
    luci_log.debug('ICB5: An error occurred while looking for cluster %s flags at path %s' % (cluname, path))
    return map

  luci_log.debug_verbose('ICB6: isClusterBusy: %s is busy: %d flags' \
      % (cluname, len(items)))
  map['busy'] = "true"
  #Ok, here is what is going on...if there is an item,
  #we need to call the ricci_bridge and get a batch report.
  #This report will tell us one of three things:
  ##1) the batch task is complete...delete ManagedSystem and render
  ##normal page
  ##2) The batch task is NOT done, so meta refresh in 5 secs and try again
  ##3) The ricci agent has no recollection of the task, so handle like 1 above
  ###
  ##Here is what we have to do:
  ##the map should have two lists:
  ##One list of non-cluster create tasks
  ##and one of cluster create task structs
  ##For each item in items, check if this is a cluster create tasktype
  ##If so, call RC, and then call stan's batch report method
  ##check for error...if error, report and then remove flag.
  ##if no error, check if complete. If not complete, report status
  ##If complete, report status and remove flag.

  for item in items:
    tasktype = item[1].getProperty(TASKTYPE)
    if tasktype == CLUSTER_ADD or tasktype == NODE_ADD:
      node_report = {}
      node_report['isnodecreation'] = True
      node_report['iserror'] = False  #Default value
      node_report['desc'] = item[1].getProperty(FLAG_DESC) 
      batch_xml = None
      ricci = item[0].split("____") #This removes the 'flag' suffix

      luci_log.debug_verbose('ICB6A: using host %s for rc for item %s' \
          % (ricci[0], item[0]))
      try:
        rc = RicciCommunicator(ricci[0])
        if not rc:
          rc = None
          raise RicciError, 'rc is None for %s' % ricci[0]
      except RicciError, e:
        rc = None
        luci_log.debug_verbose('ICB7: ricci returned error in iCB for %s: %s' \
          % (cluname, str(e)))
      except:
        rc = None
        luci_log.info('ICB8: ricci connection failed for cluster %s' % cluname)

      batch_id = None
      if rc is not None:
        try:
          batch_id = item[1].getProperty(BATCH_ID)
          luci_log.debug_verbose('ICB8A: got batch_id %s from %s' \
              % (batch_id, item[0]))
        except Exception, e:
          try:
            luci_log.debug_verbose('ICB8B: failed to get batch_id from %s: %s' \
                % (item[0], str(e)))
          except:
            luci_log.debug_verbose('ICB8C: failed to get batch_id from %s' % item[0])

        if batch_id is not None:
          try:
            batch_xml = rc.batch_report(batch_id)
            if batch_xml is not None:
              luci_log.debug_verbose('ICB8D: batch_xml for %s from batch_report is not None -- getting batch status' % batch_id)
              (creation_status, total) = batch_status(batch_xml)
              try:
                luci_log.debug_verbose('ICB8E: batch status returned (%d,%d)' \
                    % (creation_status, total))
              except:
                luci_log.debug_verbose('ICB8F: error logging batch status return')
            else:
              luci_log.debug_verbose('ICB9: batch_xml for cluster is None')
          except Exception, e:
            luci_log.debug_verbose('ICB9A: error getting batch_xml from rc.batch_report: %s' % str(e))
            creation_status = RICCI_CONNECT_FAILURE  #No contact with ricci (-1000)
            batch_xml = "bloody_failure" #set to avoid next if statement

      if rc is None or batch_id is None:
          luci_log.debug_verbose('ICB12: unable to connect to a ricci agent for cluster %s to get batch status')
          creation_status = RICCI_CONNECT_FAILURE  #No contact with ricci (-1000)
          batch_xml = "bloody_bloody_failure" #set to avoid next if statement

      if batch_xml is None:  #The job is done and gone from queue
        if redirect_message == False: #We have not displayed this message yet
          node_report['desc'] = REDIRECT_MSG
          node_report['iserror'] = True 
          node_report['errormessage'] = ""
          nodereports.append(node_report)
          redirect_message = True

        luci_log.debug_verbose('ICB13: batch job is done -- deleting %s' % item[0])
        clusterfolder.manage_delObjects(item[0])
        continue



      if creation_status < 0:  #an error was encountered
        if creation_status == RICCI_CONNECT_FAILURE:
          laststatus = item[1].getProperty(LAST_STATUS)
          if laststatus == INSTALL_TASK: #This means maybe node is rebooting
            node_report['statusindex'] = INSTALL_TASK
            node_report['statusmessage'] = RICCI_CONNECT_FAILURE_MSG + POSSIBLE_REBOOT_MESSAGE
          elif laststatus == 0:
            node_report['statusindex'] = 0
            node_report['statusmessage'] = RICCI_CONNECT_FAILURE_MSG + PRE_INSTALL
          elif laststatus == REBOOT_TASK:
            node_report['statusindex'] = REBOOT_TASK
            node_report['statusmessage'] = RICCI_CONNECT_FAILURE_MSG + PRE_CFG
          elif laststatus == SEND_CONF:
            node_report['statusindex'] = SEND_CONF
            node_report['statusmessage'] = RICCI_CONNECT_FAILURE_MSG + PRE_JOIN
          nodereports.append(node_report)
          continue
        elif creation_status == -(INSTALL_TASK):
          node_report['iserror'] = True
          (err_code, err_msg) = extract_module_status(batch_xml, INSTALL_TASK)
          node_report['errormessage'] = CLUNODE_CREATE_ERRORS[INSTALL_TASK] + err_msg
        elif creation_status == -(REBOOT_TASK):
          node_report['iserror'] = True
          (err_code, err_msg) = extract_module_status(batch_xml, REBOOT_TASK)
          node_report['errormessage'] = CLUNODE_CREATE_ERRORS[REBOOT_TASK] + err_msg
        elif creation_status == -(SEND_CONF):
          node_report['iserror'] = True
          (err_code, err_msg) = extract_module_status(batch_xml, SEND_CONF)
          node_report['errormessage'] = CLUNODE_CREATE_ERRORS[SEND_CONF] + err_msg
        elif creation_status == -(START_NODE):
          node_report['iserror'] = True
          (err_code, err_msg) = extract_module_status(batch_xml, START_NODE)
          node_report['errormessage'] = CLUNODE_CREATE_ERRORS[START_NODE]
        else:
          node_report['iserror'] = True
          node_report['errormessage'] = CLUNODE_CREATE_ERRORS[0]
        clusterfolder.manage_delObjects(item[0])
        nodereports.append(node_report)
        continue
      else:  #either batch completed successfully, or still running
        if creation_status == total:  #finished...
          map['busy'] = "true"
          node_report['statusmessage'] = "Node created successfully" + REDIRECT_MSG
          node_report['statusindex'] = creation_status
          nodereports.append(node_report)
          try:
              clusterfolder.manage_delObjects(item[0])
          except Exception, e:
              luci_log.info('ICB14: Unable to delete %s: %s' % (item[0], str(e)))
          continue
        else:
          map['busy'] = "true"
          isBusy = True
          node_report['statusmessage'] = "Node still being created"
          node_report['statusindex'] = creation_status
          nodereports.append(node_report)
          propslist = list()
          propslist.append(LAST_STATUS)
          item[1].manage_delProperties(propslist)
          item[1].manage_addProperty(LAST_STATUS,creation_status, "int")
          continue
          
    else:
      node_report = {}
      node_report['isnodecreation'] = False
      ricci = item[0].split("____") #This removes the 'flag' suffix
      rc = RicciCommunicator(ricci[0])
      finished = checkBatch(rc, item[1].getProperty(BATCH_ID))
      if finished == True:
        node_report['desc'] = item[1].getProperty(FLAG_DESC) + REDIRECT_MSG
        nodereports.append(node_report)
        try:
            clusterfolder.manage_delObjects(item[0])
        except Exception, e:
            luci_log.info('Unable to delete %s: %s' % (item[0], str(e)))
      else:
        node_report = {}
        map['busy'] = "true"
        isBusy = True
        node_report['desc'] = item[1].getProperty(FLAG_DESC)
        nodereports.append(node_report)

  if isBusy:
    part1 = req['ACTUAL_URL']
    part2 = req['QUERY_STRING']
    dex = part2.find("&busyfirst")
    if dex != (-1):
      tmpstr = part2[:dex] #This strips off busyfirst var
      part2 = tmpstr
      ###FIXME - The above assumes that the 'busyfirst' query var is at the
      ###end of the URL...
    wholeurl = part1 + "?" + part2
    #map['url'] = "5, url=" + req['ACTUAL_URL'] + "?" + req['QUERY_STRING']
    map['refreshurl'] = "5; url=" + wholeurl
    req['specialpagetype'] = "1"
  else:
    map['refreshurl'] = '5; url=\".\"'
  return map

def getClusterOS(self, rc):
	map = {}

	try:
		os_str = resolveOSType(rc.os())
		map['os'] = os_str
		map['isVirtualized'] = rc.dom0()
	except:
		# default to rhel5 if something crazy happened.
		try:
			luci_log.debug('An error occurred while attempting to get OS/Virt info for %s -- defaulting to rhel5/False' % rc.hostname())
		except:
			# this can throw an exception if the original exception
			# is caused by rc being None or stale.
			pass
		map['os'] = 'rhel5'
		map['isVirtualized'] = False
	return map

def getResourcesInfo(modelb, request):
	resList = list()
	baseurl = request['URL']

	try:
		cluname = request['clustername']
	except KeyError, e:
		try:
			cluname = request.form['clustername']
		except:
			luci_log.debug_verbose('getResourcesInfo missing cluster name')
			return resList
	except:
		luci_log.debug_verbose('getResourcesInfo missing cluster name')
		return resList

	for item in modelb.getResources():
		itemmap = {}
		itemmap['name'] = item.getName()
		itemmap['type'] = item.resource_type
		itemmap['cfgurl'] = baseurl + "?" + "clustername=" + cluname + "&resourcename=" + item.getName() + "&pagetype=" + RESOURCE_CONFIG
		itemmap['url'] = baseurl + "?" + "clustername=" + cluname + "&resourcename=" + item.getName() + "&pagetype=" + RESOURCE
		itemmap['delurl'] = baseurl + "?" + "clustername=" + cluname + "&resourcename=" + item.getName() + "&pagetype=" + RESOURCE_REMOVE
		resList.append(itemmap)
	return resList

def getResourceInfo(modelb, request):
	if not modelb:
		luci_log.debug_verbose('no modelb obj in getResourceInfo')
		return {}

	name = None
	try:
		name = request['resourcename']
	except KeyError, e:
		try:
			name = request.form['resourcename']
		except:
			pass
	except:
		pass

	if name is None:
		try:
			type = request.form['type']
			if type == 'ip':
				name = request.form['value'].strip()
		except:
			pass

	if name is None:
		luci_log.debug_verbose('getResourceInfo missing res name')
		return {}

	try:
		cluname = request['clustername']
	except KeyError, e:
		try:
			cluname = request.form['clustername']
		except:
			luci_log.debug_verbose('getResourceInfo missing cluster name')
			return {}
	except:
		luci_log.debug_verbose('getResourceInfo missing cluster name')
		return {}

	try:
		baseurl = request['URL']
	except:
		luci_log.debug_verbose('getResourceInfo missing URL')
		return {}

	for res in modelb.getResources():
		if res.getName() == name:
			try:
				resMap = {}
				resMap['name'] = res.getName()
				resMap['type'] = res.resource_type
				resMap['tag_name'] = res.TAG_NAME
				resMap['attrs'] = res.attr_hash
				resMap['cfgurl'] = baseurl + "?" + "clustername=" + cluname + "&resourcename=" + res.getName() + "&pagetype=" + RESOURCE_CONFIG
				return resMap
			except:
				continue

def delResource(self, rc, request):
	errstr = 'An error occurred in while attempting to set the cluster.conf'

	try:
		modelb = request.SESSION.get('model')
	except:
		luci_log.debug_verbose('delRes unable to extract model from SESSION')
		return errstr

	try:
		name = request['resourcename']
	except KeyError, e:
		try:
			name = request.form['resourcename']
		except:
			luci_log.debug_verbose('delRes missing resname %s' % str(e))
			return errstr + ': ' + str(e)
	except:
		luci_log.debug_verbose('delRes missing resname')
		return errstr + ': ' + str(e)

	try:
		clustername = request['clustername']
	except KeyError, e:
		try:
			clustername = request.form['clustername']
		except:
			luci_log.debug_verbose('delRes missing cluster name')
			return errstr + ': could not determine the cluster name.'

	try:
		ragent = rc.hostname()
		if not ragent:
			raise
	except:
		return errstr

	resPtr = modelb.getResourcesPtr()
	resources = resPtr.getChildren()

	found = 0
	for res in resources:
		if res.getName() == name:
			resPtr.removeChild(res)
			found = 1
			break

	if not found:
		luci_log.debug_verbose('delRes cant find res %s' % name)
		return errstr + ': the specified resource was not found.'

	try:
		conf = modelb.exportModelAsString()
		if not conf:
			raise Exception, 'model string is blank'
	except Exception, e:
		luci_log.debug_verbose('delRes: exportModelAsString failed: %s' % str(e))
		return errstr

	batch_number, result = setClusterConf(rc, str(conf))
	if batch_number is None or result is None:
		luci_log.debug_verbose('delRes: missing batch and/or result from setClusterConf')
		return errstr

	modelstr = ""
	path = CLUSTER_FOLDER_PATH + str(clustername)
	clusterfolder = self.restrictedTraverse(path)
	batch_id = str(batch_number)
	objname = str(ragent) + '____flag'
	objpath = str(path + '/' + objname)

	try:
		clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
		#Now we need to annotate the new DB object
		flag = self.restrictedTraverse(objpath)
		flag.manage_addProperty(BATCH_ID, batch_id, "string")
		flag.manage_addProperty(TASKTYPE, RESOURCE_REMOVE, "string")
		flag.manage_addProperty(FLAG_DESC, "Removing Resource \'" + request['resourcename'] + "\'", "string")
	except Exception, e:
		luci_log.debug('delRes: An error occurred while setting flag %s: %s' \
			% (objname, str(e)))
	except:
		luci_log.debug('delRes: An error occurred while setting flag %s' % objname)

	response = request.RESPONSE
	response.redirect(request['HTTP_REFERER'] + "&busyfirst=true")

def addIp(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addIp error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addIp error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addIp error: %s' % str(e))
			return None
	else:
		try:
			res = apply(Ip)
			if not res:
				raise Exception, 'apply(Ip) is None'
		except Exception, e:
			luci_log.debug_verbose('addIp error: %s' % str(e))
			return None

	if not res:
		luci_log.debug_verbose('addIp error: res is none')
		return None

	errors = list()
	try:
		addr = form['ip_address'].strip()
		if not addr:
			raise KeyError, 'ip_address is blank'
		# XXX: validate IP addr
		res.attr_hash['address'] = addr
	except KeyError, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addIp error: %s' % err)

	if 'monitorLink' in form:
		res.attr_hash['monitor_link'] = '1'
	else:
		res.attr_hash['monitor_link'] = '0'

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addFs(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addFs error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addFs error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addFs error: %s' % str(e))
			return None
	else:
		try:
			res = apply(Fs)
			if not res:
				raise Exception, 'apply(Fs) is None'
		except Exception, e:
			luci_log.debug_verbose('addFs error: %s' % str(e))
			return None

	if not res:
		luci_log.debug_verbose('addFs error: fs obj was not created')
		return None

	# XXX: sanity check these fields
	errors = list()
	try:
		name = form['resourceName'].strip()
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	try:
		mountpoint = form['mountpoint'].strip()
		res.attr_hash['mountpoint'] = mountpoint
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	try:
		device = form['device'].strip()
		res.attr_hash['device'] = device
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	try:
		options = form['options'].strip()
		res.attr_hash['options'] = options
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	try:
		fstype = form['fstype'].strip()
		res.attr_hash['fstype'] = fstype
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	try:
		fsid = form['fsid'].strip()
		res.attr_hash['fsid'] = fsid
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addFs error: %s' % err)

	if form.has_key('forceunmount'):
		res.attr_hash['force_unmount'] = '1'
	else:
		res.attr_hash['force_unmount'] = '0'

	if form.has_key('selffence'):
		res.attr_hash['self_fence'] = '1'
	else:
		res.attr_hash['self_fence'] = '0'

	if form.has_key('checkfs'):
		res.attr_hash['force_fsck'] = '1'
	else:
		res.attr_hash['force_fsck'] = '0'

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addGfs(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addGfs error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addGfs error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
			if not res:
				luci_log.debug('resource %s was not found for editing' % oldname)
				return None
		except Exception, e:
			luci_log.debug('resource %s was not found for editing: %s' \
				% (oldname, str(e)))
			return None
	else:
		try:
			res = apply(Clusterfs)
			if not res:
				raise Exception, 'apply(Clusterfs) is None'
		except Exception, e:
			luci_log.debug('addGfs error: %s' % str(e))
			return None
		except:
			luci_log.debug('addGfs error')
			return None

	# XXX: sanity check these fields
	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addGfs error: %s' % err)

	try:
		mountpoint = form['mountpoint'].strip()
		res.attr_hash['mountpoint'] = mountpoint
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addGfs error: %s' % err)

	try:
		device = form['device'].strip()
		res.attr_hash['device'] = device
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addGfs error: %s' % err)

	try:
		options = form['options'].strip()
		res.attr_hash['options'] = options
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addGfs error: %s' % err)

	try:
		fsid = form['fsid'].strip()
		res.attr_hash['fsid'] = fsid
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addGfs error: %s' % err)

	if form.has_key('forceunmount'):
		res.attr_hash['force_unmount'] = '1'
	else:
		res.attr_hash['force_unmount'] = '0'

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addNfsm(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addNfsm error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addNfsm error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addNfsm error: %s' % str(e))
			return None
	else:
		try:
			res = apply(Netfs)
		except Exception, e:
			luci_log.debug_verbose('addNfsm error: %s' % str(e))
			return None

	if not res:
		return None

	# XXX: sanity check these fields
	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)

	try:
		mountpoint = form['mountpoint'].strip()
		res.attr_hash['mountpoint'] = mountpoint
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)
		
	try:
		host = form['host'].strip()
		res.attr_hash['host'] = host
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)

	try:
		options = form['options'].strip()
		res.attr_hash['options'] = options
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)

	try:
		exportpath = form['exportpath'].strip()
		res.attr_hash['exportpath'] = exportpath 
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)

	try:
		nfstype = form['nfstype'].strip().lower()
		if nfstype != 'nfs' and nfstype != 'nfs4':
			raise KeyError, 'invalid nfs type: %s' % nfstype
		res.attr_hash['nfstype'] = nfstype
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsm error: %s' % err)

	if form.has_key('forceunmount'):
		res.attr_hash['force_unmount'] = '1'
	else:
		res.attr_hash['force_unmount'] = '0'

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addNfsc(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addNfsc error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addNfsc error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addNfsc error: %s' % str(e))
			return None
	else:
		try:
			res = apply(NFSClient)
		except:
			luci_log.debug_verbose('addNfsc error: %s' % str(e))
			return None

	if not res:
		luci_log.debug_verbose('addNfsc error: res is none')
		return None

	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsc error: %s' % err)

	try:
		target = form['target'].strip()
		res.attr_hash['target'] = target 
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsc error: %s' % err)

	try:
		options = form['options'].strip()
		res.attr_hash['options'] = options
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsc error: %s' % err)

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addNfsx(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addNfsx error: modelb is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addNfsx error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addNfsx error: %s', str(e))
			return None
	else:
		try:
			res = apply(NFSExport)
		except:
			luci_log.debug_verbose('addNfsx error: %s', str(e))
			return None

	if not res:
		luci_log.debug_verbose('addNfsx error: res is None')
		return None

	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addNfsx error: %s', err)

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addScr(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addScr error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addScr error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addScr error: %s' % str(e))
			return None
	else:
		try:
			res = apply(Script)
		except Exception, e:
			luci_log.debug_verbose('addScr error: %s' % str(e))
			return None

	if not res:
		luci_log.debug_verbose('addScr error: res is None')
		return None

	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addScr error: %s' % err)

	try:
		file = form['file'].strip()
		if not file:
			raise KeyError, 'file path is blank'
		res.attr_hash['file'] = file
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addScr error: %s' % err)

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

def addSmb(request, form=None):
	if form is None:
		form = request.form

	if not form:
		luci_log.debug_verbose('addSmb error: form is missing')
		return None

	modelb = request.SESSION.get('model')
	if not modelb:
		luci_log.debug_verbose('addSmb error: modelb is missing')
		return None

	if form.has_key('edit'):
		try:
			oldname = form['oldname'].strip()
			if not oldname:
				raise KeyError, 'oldname is blank.'
			res = getResourceForEdit(modelb, oldname)
		except Exception, e:
			luci_log.debug_verbose('addSmb error: %s' % str(e))
			return None
	else:
		try:
			res = apply(Samba)
		except Exception, e:
			luci_log.debug_verbose('addSmb error: %s' % str(e))
			return None

	if not res:
		luci_log.debug_verbose('addSmb error: res is None')
		return None

	errors = list()
	try:
		name = form['resourceName'].strip()
		if not name:
			raise KeyError, 'resourceName is blank'
		res.attr_hash['name'] = name
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addSmb error: %s' % err)

	try:
		workgroup = form['workgroup'].strip()
		res.attr_hash['workgroup'] = workgroup
	except Exception, e:
		err = str(e)
		errors.append(err)
		luci_log.debug_verbose('addSmb error: %s' % err)

	if len(errors) > 1:
		return [None, None, errors]
	return [res, modelb, None]

resourceAddHandler = {
	'ip': addIp,
	'fs': addFs,
	'gfs': addGfs,
	'nfsm': addNfsm,
	'nfsx': addNfsx,
	'nfsc': addNfsx,
	'scr': addScr,
	'smb': addSmb
}

def resolveClusterChanges(self, clusterName, modelb):
	try:
		mb_nodes = modelb.getNodes()
		if not mb_nodes or not len(mb_nodes):
			raise Exception, 'node list is empty'
	except Exception, e:
		luci_log.debug_verbose('no model builder nodes found for %s: %s' \
				% (str(e), clusterName))
		return 'Unable to find cluster nodes for %s' % clusterName

	try:
		cluster_node = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		if not cluster_node:
			raise Exception, 'cluster node is none'
	except Exception, e:
		luci_log.debug('cant find cluster node for %s: %s'
			% (clusterName, str(e)))
		return 'Unable to find an entry for %s in the Luci database.' % clusterName

	try:
		db_nodes = map(lambda x: x[0], cluster_node.objectItems('Folder'))
		if not db_nodes or not len(db_nodes):
			raise
	except:
		# Should we just create them all? Can this even happen?
		return 'Unable to find database entries for any nodes in ' + clusterName

	same_host = lambda x, y: x == y or x[:len(y) + 1] == y + '.' or y[:len(x) + 1] == x + '.'

	# this is a really great algorithm.
	missing_list = list()
	new_list = list()
	for i in mb_nodes:
		for j in db_nodes:
			f = 0
			if same_host(i, j):
				f = 1
				break
		if not f:
			new_list.append(i)

	for i in db_nodes:
		for j in mb_nodes:
			f = 0
			if same_host(i, j):
				f = 1
				break
		if not f:
			missing_list.append(i)

	messages = list()
	for i in missing_list:
		cluster_node.delObjects([i])
		## or alternately
		#new_node = cluster_node.restrictedTraverse(i)
		#setNodeFlag(self, new_node, CLUSTER_NODE_NOT_MEMBER)
		messages.append('Node \"' + i + '\" is no longer in a member of cluster \"' + clusterName + '.\". It has been deleted from the management interface for this cluster.')

	new_flags = CLUSTER_NODE_NEED_AUTH | CLUSTER_NODE_ADDED
	for i in new_list:
		try:
			cluster_node.manage_addFolder(i, '__luci__:csystem:' + clusterName)
			new_node = cluster_node.restrictedTraverse(i)
			setNodeFlag(self, new_node, new_flags)
			messages.append('A new node, \"' + i + ',\" is now a member of cluster \"' + clusterName + '.\" It has added to the management interface for this cluster, but you must authenticate to it in order for it to be fully functional.')
		except:
			messages.append('A new node, \"' + i + ',\" is now a member of cluster \"' + clusterName + ',\". but has not added to the management interface for this cluster as a result of an error creating the database entry.')
	
	return messages

def addResource(self, request, modelb, res):
	clustername = modelb.getClusterName()
	if not clustername:
		raise Exception, 'cluster name from modelb.getClusterName() is blank'

	rc = getRicciAgent(self, clustername)
	if not rc:
		raise Exception, 'Unable to find a ricci agent for the %s cluster' % clustername

	modelb.getResourcesPtr().addChild(res)

	try:
		conf = modelb.exportModelAsString()
		if not conf:
			raise Exception, 'model string for %s is blank' % clustername
	except Exception, e:
		luci_log.debug_verbose('addResource: exportModelAsString err: %s' % str(e))
		return 'An error occurred while adding this resource'

	try:
		ragent = rc.hostname()
		if not ragent:
			luci_log.debug_verbose('missing hostname')
			raise Exception, 'unknown ricci agent hostname'
		luci_log.debug_verbose('SENDING NEW CLUSTER CONF: %s' % conf)
		batch_number, result = setClusterConf(rc, str(conf))
		if batch_number is None or result is None:
			luci_log.debug_verbose('missing batch_number or result')
			raise Exception, 'batch_number or results is None from setClusterConf'
	except Exception, e:
		return 'An error occurred while propagating the new cluster.conf: %s' % str(e)

	path = str(CLUSTER_FOLDER_PATH + clustername)
	clusterfolder = self.restrictedTraverse(path)
	batch_id = str(batch_number)
	objname = str(ragent + '____flag')
	objpath = str(path + '/' + objname)

	try:
		clusterfolder.manage_addProduct['ManagedSystem'].addManagedSystem(objname)
		#Now we need to annotate the new DB object
		flag = self.restrictedTraverse(objpath)
		flag.manage_addProperty(BATCH_ID, batch_id, "string")
		flag.manage_addProperty(TASKTYPE, RESOURCE_ADD, "string")

		if type != 'ip':
			flag.manage_addProperty(FLAG_DESC, "Creating New Resource \'" + res.attr_hash['name'] + "\'", "string")
		else:
			flag.manage_addProperty(FLAG_DESC, "Creating New Resource \'" + res.attr_hash['address'] + "\'", "string")
	except Exception, e:
		try:
			luci_log.info('Unable to create flag %s: %s' % (objpath, str(e)))
		except:
			pass

	response = request.RESPONSE
	response.redirect(request['HTTP_REFERER'] + "&busyfirst=true")

def getResourceForEdit(modelb, name):
	resPtr = modelb.getResourcesPtr()
	resources = resPtr.getChildren()

	for res in resources:
		if res.getName() == name:
			resPtr.removeChild(res)
			return res

	luci_log.debug_verbose('unable to find resource \"%s\"' % name)
	raise KeyError, name

def appendModel(request, model):
	try:
		request.SESSION.set('model', model)
	except:
		luci_log.debug_verbose('Appending model to request failed')
		return 'An error occurred while storing the cluster model.' 

def resolve_nodename(self, clustername, nodename):
	path = str(CLUSTER_FOLDER_PATH + clustername)

	try:
		clusterfolder = self.restrictedTraverse(path)
		objs = clusterfolder.objectItems('Folder')
	except Exception, e:
		luci_log.info('resolve_nodename failed for %s/%s: %s' \
			% (nodename, clustername, str(e)))

	for obj in objs:
		if obj[0].find(nodename) != (-1):
			return obj[0]

	luci_log.info('resolve_nodename failed for %s/%s' % (nodename, clustername))
	return None

def noNodeFlagsPresent(self, nodefolder, flagname, hostname):
	try:
		items = nodefolder.objectItems('ManagedSystem')
	except:
		luci_log.debug('An error occurred while trying to list flags for cluster ' + nodefolder[0])
		return None

	for item in items:
		if item[0] != flagname:
			continue

		#a flag already exists... try to delete it
		try:
			rc = RicciCommunicator(hostname)
		except RicciError, e:
			luci_log.info('Unable to connect to the ricci daemon: %s' % str(e))
			return None

		if not rc.authed():
			try:
				snode = getStorageNode(self, hostname)
				setNodeFlag(snode, CLUSTER_NODE_NEED_AUTH)
			except:
				pass
			luci_log.info('Node %s is not authenticated' % item[0])
			return None

		finished = checkBatch(rc, item[1].getProperty(BATCH_ID))
		if finished == True:
			try:
				nodefolder.manage_delObjects(item[0])
			except Exception, e:
				luci_log.info('manage_delObjects for %s failed: %s' \
					% (item[0], str(e)))
				return None
			return True
		else:
			#Not finished, so cannot remove flag
			return False

	return True

def getModelBuilder(rc, isVirtualized):
	try:
		cluster_conf_node = getClusterConf(rc)
		if not cluster_conf_node:
			raise
	except:
		luci_log.debug('unable to get cluster_conf_node in getModelBuilder')
		return None

	try:
		modelb = ModelBuilder(0, None, None, cluster_conf_node)
	except Exception, e:
		try:
			luci_log.debug('An error occurred while trying to get modelb for conf \"%s\": %s' % (cluster_conf_node.toxml(), str(e)))
		except:
			pass

	modelb.setIsVirtualized(isVirtualized)
	return modelb
