import string
import re
import sys
import os
from AccessControl import getSecurityManager
from ZPublisher import HTTPRequest
import xml.dom
import cgi

from ricci_defines import *
from ricci_bridge import getClusterConf
from ricci_communicator import RicciCommunicator
from ricci_communicator import CERTS_DIR_PATH
from clusterOS import resolveOSType
from conga_constants import *
from LuciSyslog import LuciSyslog, LuciSyslogError

try:
	luci_log = LuciSyslog()
except LuciSyslogError, e:
	pass

def siteIsSetup(self):
	try:
		if os.path.isfile(CERTS_DIR_PATH + 'privkey.pem') and os.path.isfile(CERTS_DIR_PATH + 'cacert.pem'):
			return True
	except: pass
	return False

def strFilter(regex, replaceChar, str):
	return re.sub(regex, replaceChar, str)

def validateDelSystem(self, request):
	errors = list()
	messages = list()

	if '__SYSTEM' in request.form:
		sysNames = request.form['__SYSTEM']
		for i in sysNames:
			if i == '':
				continue
			dsResult = delSystem(self, i)
			if dsResult:
				errors.append(dsResult)
			else:
				messages.append('Removed storage system \"' + i + '\" successfully')

	if '__CLUSTER' in request.form:
		cluNames = request.form['__CLUSTER']
		for i in cluNames:
			if i == '':
				continue
			dcResult = delCluster(self, i)
			if dcResult:
				errors.append(dcResult)
			else:
				messages.append('Removed cluster \"' + i + '\" successfully')

	if len(errors) > 0:
		retCode = False
	else:
		retCode = True

	return (retCode, {'messages': messages, 'errors': errors })

def validateDelUser(self, request):
	errors = list()
	messages = list()

	if not 'deluserId' in request.form:
		return (False, {'errors': [ 'No User Id given' ] })
	
	userId = request.form['deluserId']

	try:
		user = self.portal_membership.getMemberById(userId)
	except:
		return (False, {'errors': [ 'No such user: \"' + userId + '\"' ] })

	for i in getClusters(self):
		try:
			i[1].manage_delLocalRoles([userId])
		except:
			errors.append('Error deleting roles from cluster \"' + i[0] + '\" for user \"' + userId + '\"')

	for i in getStorage(self):
		try:
			i[1].manage_delLocalRoles([userId])
		except:
			errors.append('Error deleting roles from storage system \"' + i[0] + '\" for user \"' + userId + '\"')

	try:
		self.acl_users.userFolderDelUsers([userId])
	except:
		errors.append('Unable to delete user \"' + userId + '\"')
		return (False, {'errors': errors })

	messages.append('User \"' + userId + '\" has been deleted')
	return (True, {'errors': errors, 'messages': messages })

def validateAddUser(self, request):
	messages = list()

	if not 'newUserName' in request.form:
		return (False, { 'errors': ['No user name given'] })
	if not 'newPassword' in request.form:
		return (False, { 'errors': ['No password given'] })
	if not 'newPasswordConfirm' in request.form:
		return (False, {'errors': ['You didn\'t confirm the password'] })

	user = request.form['newUserName']

	if self.portal_membership.getMemberById(user):
		return (False, {'errors': ['The user \"' + user + '\" already exists']})

	passwd = request.form['newPassword']
	pwconfirm = request.form['newPasswordConfirm']

	if passwd != pwconfirm:
		return (False, {'errors': ['The passwords do not match']})

	try:
		self.portal_registration.addMember(user, passwd, properties = { 'username': user, 'password': passwd, 'confirm': passwd, 'roles': ['Member'], 'domains':[], 'email': user + '@example.com' })
	except:
		return (False, {'errors': [ 'Unable to add new user \"' + user + '\"' ] })

	if not self.portal_membership.getMemberById(user):
		return (False, {'errors': [ 'Unable to add new user \"' + user + '\"'] })

	messages.append('Added new user \"' + user + '\" successfully')
	return (True, {'messages': messages, 'params': { 'user': user }})

def nodeUnauth(nodeList):
	for i in nodeList:
		try:
			if i['prev_auth'] != True:
				host = i['host']
				rc = RicciCommunicator(host)
				rc.unauth()
				i['cur_auth'] = False
		except:
			pass

def nodeAuth(cluster, host, passwd):
	messages = list()
	systemName = host
	os_str = 'rhel5'

	try:
		rc = RicciCommunicator(host)
		if not rc:
			luci_log.debug_verbose('nodeAuth0: rc is None')
			raise Exception, 'unknown error'
	except Exception, e:
		try:
			error = 'Ricci connection to %s failed: %s' % (host, str(e))
		except:
			error = 'Ricci connection to %s failed' % host
		luci_log.debug_verbose('nodeAuth1: rc failed: %s' % error)

		return { 'host': host, 'ricci_host': host, 'errors': error, 'cur_auth': False, 'os': os_str }

	if rc.authed():
		prevAuth = True
		messages.append('Luci is already authenticated to %s -- not checking password' % host)
	else:
		prevAuth = False
		if not passwd:
			return { 'host': host, 'ricci_host': systemName, 'prev_auth': False, 'cur_auth': False, 'os': os_str }
		else:
			try:
				rc.auth(passwd)
			except:
				pass

	if rc.authed():
		try:
			os_str = resolveOSType(rc.os())
			if not os_str:
				raise
		except:
			os_str = "rhel5"  #Backup plan in case all is almost lost...

		systemName = rc.system_name()
		if systemName[:9] == 'localhost' or systemName[:5] == '127.0':
			systemName = host
		node = { 'host': host, 'ricci_host': systemName, 'prev_auth': prevAuth, 'cur_auth': True, 'os': os_str }

		cluster_info = rc.cluster_info()
		if cluster and ((not cluster_info) or (cluster_info[0] != cluster)):
			node['errors'] = 'Node \"' + host + '\" is reporting it is not a member of cluster \"' + cluster + '\"'
			if cluster_info and cluster_info[0]:
				node['errors'] += ' and that it is a member of cluster \"' + cluster_info[0] + '\"'
		if not cluster and cluster_info and cluster_info[0]:
			node['errors'] = 'Node \"' + host + '\" reports it is a member of cluster \"' + cluster_info[0] + '\"'
		return node

	error = 'Unable to authenticate to the ricci agent on \"' + host + '\"'
	return { 'host': host, 'ricci_host': systemName, 'prev_auth': False , 'cur_auth': False, 'errors': error, 'os': os_str }

def validateAddClusterInitial(self, request, must_complete=True):
	errors = list()
	messages = list()
	newNodeList = list()
	nodeHash = {}
	rnodeHash = {}

	try:
		sysData = request.form['__SYSTEM0']
		if not sysData or len(sysData) < 2:
			raise
	except:
		return (False, { 'errors': [ 'At least one system and its root password must be given' ] })

	try:
		rc = RicciCommunicator(sysData[0])
		if not rc:
			raise Exception, 'unknown error'
	except Exception, e:
		return (False, { 'errors': [ 'Unable to establish a connection to the Ricci agent on %s: %s' % (sysData[0], str(e)) ] })

	prevAuth = 0
	if not rc.authed():
		try:
			rc.auth(sysData[1])
		except: pass
		if not rc.authed():
			return (False, { 'errors': [ 'Unable to authenticate to the Ricci agent on \"' + sysData[0] + '\"' ] })
	else:
		prevAuth = 1

	try:
		cluster_info = rc.cluster_info()
	except:
		cluster_info = None

	os_str = resolveOSType(rc.os())
	if not os_str:
		os_str = "rhel5"  #Backup plan in case all is almost lost...

	if not cluster_info or not cluster_info[0]:
		if not prevAuth:
			rc.unauth()
		if not cluster_info:
			errmsg = 'An error occurred while attempting to retrieve the cluster.conf file for \"' + sysData[0] + '\"'
		else:
			errmsg = '\"' + sysData[0] + '\" is not a member of a cluster'
		return (False, { 'errors': [ errmsg ] })

	clusterName = cluster_info[0]
	cluConf = getClusterConf(rc)
	if cluConf:
		nodeList = getClusterConfNodes(cluConf)

	if not cluConf or not nodeList or len(nodeList) < 1:
		if not prevAuth:
			rc.unauth()
		return (False, { 'errors': [ 'Error retrieving member nodes for cluster \"' + clusterName + '\"' ] })

	systemName = rc.system_name()
	if systemName[:9] == 'localhost':
		systemName = sysData[0]

	node = { 'host': rc.hostname(), 'ricci_host': systemName, 'prev_auth': prevAuth, 'cur_auth': rc.authed(), 'os': os_str }
	nodeHash[sysData[0]] = node
	rnodeHash[systemName] = node
	newNodeList.append(node)

	if 'allSameCheckBox' in request.form:
		passwd = sysData[1]
	else:
		passwd = None
		
	for i in nodeList:
		node = nodeAuth(clusterName, i, passwd)
		if 'messages' in node:
			messages.extend(node['messages'])
		if node['host'] in nodeHash or node['ricci_host'] in rnodeHash:
			continue
		nodeHash[node['host']] = node
		if 'ricci_host' in node:
			rnodeHash[node['ricci_host']] = node

		if 'errors' in node:
			errors.append(node['errors'])
			node['errors'] = True
		newNodeList.append(node)

	sfn = lambda x, y: \
		x['cur_auth'] - y['cur_auth'] or (('errors' in y) - ('errors' in x)) 
	newNodeList.sort(sfn)

	if must_complete == True:
		dfn = lambda x: not 'cur_auth' in x or x['cur_auth'] != True
	else:
		dfn = lambda x: False

	cluster_properties = {
		'clusterName': clusterName,
		'nodeList': newNodeList,
		'nodeHash': nodeHash,
		'rnodeHash': rnodeHash,
		'isComplete': len(filter(dfn, newNodeList)) == 0
	}

	if len(errors) < len(nodeList):
		cluster_properties['redirect'] = HOMEBASE_ADD_CLUSTER

	return (len(errors) < 1,
		{'messages': messages, 'errors': errors, 'requestResults': cluster_properties })

def validateAddCluster(self, request, must_complete=True):
	errors = list()
	messages = list()
	requestResults = None
	nodeList = None

	try:
		sessionData = request.SESSION.get('checkRet')
		requestResults = sessionData['requestResults']
	except Exception, e:
		luci_log.debug_verbose('VAC0: error getting session obj: %s' % str(e))
		try:
			clusterName = request.form['clusterName']
		except:
			clusterName = ''

		try:
			nodeList = requestResults['nodeList']
			luci_log.debug_verbose('VAC1: unauth to node list')
			nodeUnauth(nodeList)
		except:
			pass

		return (False, { 'errors': [ 'A data integrity error has occurred. Please attempt adding the cluster again.' ], 'requestResults': { 'clusterName': clusterName, 'isComplete': False, 'nodeList': [], 'redirect': HOMEBASE_ADD_CLUSTER_INITIAL } })
		
	try:
		clusterName = request.form['clusterName']
		if not clusterName:
			raise Exception, 'no cluster name was found'
	except Exception, e:
		luci_log.debug_verbose('VAC2: no cluser name found: %s', str(e))
		return (False, { 'errors': ['No cluster name was given.'], 'requestResults': requestResults })

	try:
		nodeList = requestResults['nodeList']
		if not nodeList or len(nodeList) < 1:
			raise Exception, 'no node list found'
	except Exception, e:
		luci_log.debug_verbose('VAC3: no nodeList found: %s', str(e))
		return (False, { 'errors': ['No cluster nodes were given.'], 'requestResults': requestResults })

	try:
		nodeHash = requestResults['nodeHash']
	except:
		nodeHash = {}

	try:
		rnodeHash = requestResults['rnodeHash']
	except:
		rnodeHash = {}

	# This should never fail
	try:
		numStorage = int(request.form['numStorage'])
		if numStorage != len(nodeList):
			raise Exception, 'numstorage != len(nodelist)'
	except Exception, e:
		try:
			requestResults['isComplete'] = False
			luci_log.debug_verbose('VAC4: error: %s' % str(e))
		except:
			pass

		nodeUnauth(nodeList)
		return (False, {
				'errors': [ 'Unknown number of nodes entered' ],
				'requestResults': requestResults })

	i = 0
	while i < numStorage:
		sysData = request.form['__SYSTEM' + str(i)]
		if not sysData:
			i += 1
			continue

		oldNode = None
		node = nodeAuth(clusterName, sysData[0], sysData[1])
		if node['host'] in nodeHash:
			oldNode = nodeHash[node['host']]
		elif 'ricci_host' in node and node['ricci_host'] in rnodeHash:
			oldNode = rnodeHash[node['ricci_host']]
		elif not oldNode:
			for k in nodeHash.keys():
				if node['host'][:len(k) + 1] == k + '.':
					oldNode = nodeHash[k]
		elif not oldNode:
			for k in rnodeHash.keys():
				if node['host'][:len(k) + 1] == k + '.':
					oldNode = rnodeHash[k]

		if not oldNode:
			luci_log.debug_verbose('VAC5: node %s not found', sysData[0])
			nodeUnauth(nodeList)
			return (False, { 'errors': [ 'A data integrity error has occurred. Please attempt adding the cluster again.' ], 'requestResults': { 'clusterName': clusterName, 'nodeList': nodeList, 'isComplete': False, 'redirect': HOMEBASE_ADD_CLUSTER_INITIAL } })

		if oldNode['host'] != node['host']:
			del nodeHash[oldNode['host']]
			oldNode['host'] = node['host']
			nodeHash[node['host']] = oldNode

		if 'ricci_host' in node and (not 'ricci_host' in oldNode or node['ricci_host'] != oldNode['ricci_host']):
			if oldNode['ricci_host'] in rnodeHash:
				del rnodeHash[oldNode['ricci_host']]
				oldNode['ricci_host'] = node['ricci_host']
				rnodeHash[node['ricci_host']] = oldNode

		oldNode['cur_auth'] = node['cur_auth']
		if 'errors' in node:
			errors.append(node['errors'])
			oldNode['errors'] = True
		i += 1

	if must_complete == True:
		dfn = lambda x: not 'cur_auth' in x or x['cur_auth'] != True
	else:
		dfn = lambda x: False

	clusterComplete = len(filter(dfn, nodeList)) == 0

	if clusterComplete:
		err = manageCluster(self, clusterName, nodeList)
		if err:
			errors.append(err)
		else:
			messages.append('Cluster \"' + clusterName + '\" has been added to the Luci management interface.') 
	else:
		sfn = lambda x, y: \
			x['cur_auth'] - y['cur_auth'] or (('errors' in y) - ('errors' in x)) 
		nodeList.sort(sfn)

	ret = { 'messages': messages, 'errors': errors }

	if len(errors) > 0 or not clusterComplete:
		ret['requestResults'] = {
			'clusterName': clusterName,
			'nodeList': nodeList,
			'nodeHash': nodeHash,
			'rnodeHash': rnodeHash,
			'isComplete': clusterComplete
		}
	else:
		ret['requestResults'] = {
			'redirect': HOMEBASE_ADD_CLUSTER_INITIAL,
			'clusterName': clusterName,
			'isComplete': True
		}

	return (len(errors) < 1, ret)

def validateAddSystem(self, request):
	errors = list()
	messages = list()

	try:
		numStorage = request.form['numStorage']
	except:
		return (False, { 'errors': ['Unknown number of systems entered'] })

	i = 0
	while i < numStorage:
		try:
			sysData = request.form['__SYSTEM' + str(i)]
		except:
			break

		if len(sysData) == 2 and sysData[0] != '' and sysData[1] != '':
			csResult = createSystem(self, sysData[0], sysData[1])

			if csResult:
				errors.append(csResult)
			else:
				messages.append('Added storage system \"' + sysData[0] + '\" successfully')
		i += 1

	if len(errors) > 0:
		returnCode = False
	else:
		returnCode = True

	return (returnCode, {'errors': errors, 'messages': messages})

def validatePerms(self, request):
	userId = None
	messages = list()
	errors = list()

	try:
		userId = request.form['userList']
	except:
		return (False, {'errors': [ 'No user specified' ], 'params': { 'user': userId }})

	user = self.portal_membership.getMemberById(userId)
	if not user:
		return (False, {'errors': [ 'Invalid user specified' ], 'params': { 'user': userId }})

	userId = user.getUserId()

	clusters = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/objectItems')('Folder')
	if not '__CLUSTER' in request.form:
		for i in clusters:
			try:
				if user.has_role('View', i[1]):
					roles = list(i[1].get_local_roles_for_userid(userId))
					roles.remove('View')

					if roles:
						i[1].manage_setLocalRoles(userId, roles)
					else:
						i[1].manage_delLocalRoles([userId])
					messages.append('Removed permission for ' + userId + ' for cluster ' + i[0])
			except:
				errors.append('Failed to remove permission for ' + userId + ' for cluster ' + i[0])
	else:
		for i in clusters:
			if i[0] in request.form['__CLUSTER']:
				try:
					if not user.has_role('View', i[1]):
						roles = list(i[1].get_local_roles_for_userid(userId))
						roles.append('View')
						i[1].manage_setLocalRoles(userId, roles)
						messages.append('Added permission for ' + userId + ' for cluster ' + i[0])
				except:
						errors.append('Failed to add permission for ' + userId + ' for cluster ' + i[0])
			else:
				try:
					if user.has_role('View', i[1]):
						roles = list(i[1].get_local_roles_for_userid(userId))
						roles.remove('View')

						if roles:
							i[1].manage_setLocalRoles(userId, roles)
						else:
							i[1].manage_delLocalRoles([userId])

						messages.append('Removed permission for ' + userId + ' for cluster ' + i[0])
				except:
						errors.append('Failed to remove permission for ' + userId + ' for cluster ' + i[0])

	storage = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/objectItems')('Folder')
	if not '__SYSTEM' in request.form:
		for i in storage:
			try:
				if user.has_role('View', i[1]):
					roles = list(i[1].get_local_roles_for_userid(userId))
					roles.remove('View')

					if roles:
						i[1].manage_setLocalRoles(userId, roles)
					else:
						i[1].manage_delLocalRoles([userId])
					messages.append('Removed permission for ' + userId + ' for ' + i[0])
			except:
				errors.append('Failed to remove permission for ' + userId + ' for ' + i[0])
	else:
		for i in storage:
			if i[0] in request.form['__SYSTEM']:
				try:
					if not user.has_role('View', i[1]):
						roles = list(i[1].get_local_roles_for_userid(userId))
						roles.append('View')
						i[1].manage_setLocalRoles(userId, roles)
						messages.append('Added permission for ' + userId + ' for system ' + i[0])
				except:
						errors.append('Failed to add permission for ' + userId + ' for system ' + i[0])
			else:
				try:
					if user.has_role('View', i[1]):
						roles = list(i[1].get_local_roles_for_userid(userId))
						roles.remove('View')

						if roles:
							i[1].manage_setLocalRoles(userId, roles)
						else:
							i[1].manage_delLocalRoles([userId])

						messages.append('Removed permission for ' + userId + ' for system ' + i[0])
				except:
						errors.append('Failed to remove permission for ' + userId + ' for system ' + i[0])

	if len(errors) > 0:
		returnCode = False
	else:
		returnCode = True

	return (returnCode, {'errors': errors, 'messages': messages, 'params': {'user': userId }})

def validateAuthenticate(self, request):
	errors = list()
	messages = list()

	try:
		numStorage = int(request.form['numStorage'])
	except:
		return (False, {'errors': [ 'Unknown number of nodes entered']})

	i = 0
	while i < numStorage:
		sysData = request.form['__SYSTEM' + str(i)]
		if not sysData or len(sysData) < 2 or not sysData[0] or not sysData[1]:
			i += 1
			continue

		host = str(sysData[0])
		passwd = str(sysData[1])

		try:
			rc = RicciCommunicator(sysData[0])
			if rc is None:
				raise Exception, 'unknown error'
		except Exception, e:
			errors.append('Unable to contact the ricci agent for %s: %s' \
				% (sysData[0], str(e)))
			i += 1
			continue

		if rc.authed():
			messages.append(host + ' is already authenticated.')
		else:
			try:
				rc.auth(passwd)
			except:
				errors.append('Error authenticating to the ricci agent on ' + host)
				i += 1
				continue

			if not rc.authed():
				errors.append('Error authenticating to the ricci agent on ' + host)
			else:
				messages.append(host + ' was successfully authenticated.')

			if rc.authed():
				try:
					delNodeFlag(self, getStorageNode(self, host), CLUSTER_NODE_NEED_AUTH)
				except:
					pass

				try:
					delNodeFlag(self, getClusterNode(self, host, rc.cluster_info()[0]), CLUSTER_NODE_NEED_AUTH)
				except:
					pass
		i += 1
			
	return (len(errors) > 0, {'errors': errors, 'messages': messages })

formValidators = [
	validateAddUser,
	validateAddSystem,
	validatePerms,
	validateDelUser,
	validateDelSystem,
	validateAddCluster,
	validateAddClusterInitial,
	validateAuthenticate
]

def userAuthenticated(self):
	if (isAdmin(self) or getSecurityManager().getUser().has_role('Authenticated', self.restrictedTraverse(PLONE_ROOT))):
		return True

	return False

def isAdmin(self):
	try:
		return getSecurityManager().getUser().has_role('Owner', self.restrictedTraverse(PLONE_ROOT))
	except:
		pass
	return False

def userIsAdmin(self, userId):
	try:
		return self.portal_membership.getMemberById(userId).has_role('Owner', self.restrictedTraverse(PLONE_ROOT))
	except:
		pass
	return False

def homebaseControlPost(self, request):
	try:
		sessionData = request.SESSION.get('checkRet')
	except:
		sessionData = None

	if 'ACTUAL_URL' in request:
		url = request['ACTUAL_URL']
	else:
		url = '.'

	if 'pagetype' in request.form:
		pagetype = int(request.form['pagetype'])
	else:
		try: request.SESSION.set('checkRet', {})
		except: pass
		return homebasePortal(self, request, '.', '0')

	try:
		validatorFn = formValidators[pagetype - 1]
	except:
		try: request.SESSION.set('checkRet', {})
		except: pass
		return homebasePortal(self, request, '.', '0')

	if validatorFn == validateAddClusterInitial or validatorFn == validateAddCluster:
		ret = validatorFn(self, request, must_complete=False)
	else:
		ret = validatorFn(self, request)
	params = None

	if 'params' in ret[1]:
		params = ret[1]['params']

	if 'requestResults' in ret[1]:
		requestResults = ret[1]['requestResults']

		if 'redirect' in requestResults:
			pagetype = requestResults['redirect']
			request['pagetype'] = pagetype
			request.form['pagetype'] = pagetype

	request.SESSION.set('checkRet', ret[1])
	return homebasePortal(self, request, url, pagetype, params)

def homebaseControl(self, request):
	if request.REQUEST_METHOD == 'POST':
		return homebaseControlPost(self, request)

	try:
		request.SESSION.set('checkRet', {})
	except:
		pass

	if 'ACTUAL_URL' in request:
		url = request['ACTUAL_URL']
	else:
		url = '.'

	if 'pagetype' in request:
		pagetype = request['pagetype']
	else:
		pagetype = '0'

	return homebasePortal(self, request, url, pagetype)

def homebasePortal(self, request=None, url=None, pagetype=None, params=None):
	ret = {}
	temp = list()
	index = 0
	cur = None

	pagetype = str(pagetype)

#
# The Add System page
#

	try:
		if havePermAddStorage(self):
			addSystem = {}
			addSystem['Title'] = 'Add a System'
			addSystem['absolute_url'] = url + '?pagetype=' + HOMEBASE_ADD_SYSTEM
			addSystem['Description'] = 'Add a system to the Luci storage management interface.'
			if pagetype == HOMEBASE_ADD_SYSTEM:
				cur = addSystem
				addSystem['currentItem'] = True
				ret['curIndex'] = index
			else:
				addSystem['currentItem'] = False
			index += 1
			temp.append(addSystem)
	except: pass

	# Initial add cluster page
	try:
		if pagetype == HOMEBASE_ADD_CLUSTER:
			raise
		if havePermAddCluster(self):
			addCluster = {}
			addCluster['Title'] = 'Add an Existing Cluster'
			addCluster['absolute_url'] = url + '?pagetype=' + HOMEBASE_ADD_CLUSTER_INITIAL
			addCluster['Description'] = 'Add an existing cluster to the Luci cluster management interface.'
			if pagetype == HOMEBASE_ADD_CLUSTER_INITIAL:
				addCluster['currentItem'] = True
				ret['curIndex'] = index
				cur = addCluster
			else:
				addCluster['currentItem'] = False
			index += 1
			temp.append(addCluster)
	except: pass

	# Add cluster - screen 2
	try:
		if pagetype != HOMEBASE_ADD_CLUSTER:
			raise
		if havePermAddCluster(self):
			addCluster = {}
			addCluster['Title'] = 'Add an Existing Cluster'
			addCluster['absolute_url'] = url + '?pagetype=' + HOMEBASE_ADD_CLUSTER
			addCluster['Description'] = 'Add an existing cluster to the Luci cluster management interface.'
			if pagetype == HOMEBASE_ADD_CLUSTER:
				addCluster['currentItem'] = True
				ret['curIndex'] = index
				cur = addCluster
			else:
				addCluster['currentItem'] = False
			index += 1
			temp.append(addCluster)
	except: pass

#
# The Remove Storage System page
#

	try:
		if (havePermRemStorage(self) and havePermRemCluster(self) and (getStorage(self) or getClusters(self))):
			remSystem = {}
			remSystem['Title'] = 'Manage Systems'
			remSystem['absolute_url'] = url + '?pagetype=' + HOMEBASE_DEL_SYSTEM
			remSystem['Description'] = 'Update or remove storage systems and clusters.'
			if pagetype == HOMEBASE_DEL_SYSTEM:
				remSystem['currentItem'] = True
				ret['curIndex'] = index
				cur = remSystem
			else:
				remSystem['currentItem'] = False
			index += 1
			temp.append(remSystem)
	except: pass

#
# Add a Luci user.
#

	try:
		if havePermAddUser(self):
			addUser = {}
			addUser['Title'] = 'Add a User'
			addUser['absolute_url'] = url + '?pagetype=' + HOMEBASE_ADD_USER
			addUser['Description'] = 'Add a user to the Luci interface.'
			if pagetype == HOMEBASE_ADD_USER:
				addUser['currentItem'] = True
				ret['curIndex'] = index
				cur = addUser
			else:
				addUser['currentItem'] = False
			index += 1
			temp.append(addUser)
	except: pass

#
# Delete a Luci user
#

	try:
		if (self.portal_membership.listMembers() and havePermDelUser(self)):
			delUser = {}
			delUser['Title'] = 'Delete a User'
			delUser['absolute_url'] = url + '?pagetype=' + HOMEBASE_DEL_USER
			delUser['Description'] = 'Delete a Luci user.'
			if pagetype == HOMEBASE_DEL_USER:
				delUser['currentItem'] = True
				ret['curIndex'] = index
				cur = delUser
			else:
				delUser['currentItem'] = False
			index += 1
			temp.append(delUser)
	except: pass

#
# The User Permissions Page
#
	try:
		if (havePermEditPerms(self) and self.portal_membership.listMembers() and (getStorage(self) or getClusters(self))):
			userPerm = {}
			userPerm['Title'] = 'User Permissions'
			userPerm['absolute_url'] = url + '?pagetype=' + HOMEBASE_PERMS
			userPerm['Description'] = 'Set permissions for Luci users.'
			if pagetype == HOMEBASE_PERMS:
				userPerm['currentItem'] = True
				ret['curIndex'] = index
				cur = userPerm
			else:
				userPerm['currentItem'] = False
			index += 1
			temp.append(userPerm)
	except: pass

	if not 'curIndex' in ret:
		ret['curIndex'] = 0

	if cur and 'absolute_url' in cur and params:
		cur['base_url'] = cur['absolute_url']
		for i in params:
			cur['absolute_url'] += '&' + cgi.escape(i) + '=' + cgi.escape(params[i])
	elif cur and 'absolute_url' in cur:
		cur['base_url'] = cur['absolute_url']
	else:
		cur = {}
		cur['base_url'] = '#'

	ret['children'] = temp
	return ret

def getClusterSystems(self, clusterName):
	if isAdmin(self):
		return self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName + '/objectItems')('Folder')

	try:
		i = getSecurityManager().getUser()
		if not i:
			raise
	except:
		return None

	csystems = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName + '/objectItems')('Folder')
	if not csystems:
		return None

	allowedCSystems = list()
	for c in csystems:
		if i.has_role('View', c[1]):
			allowedCSystems.append(c)
	return (c)

def getClusters(self):
	if isAdmin(self):
		return self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/objectItems')('Folder')
	try:
		i = getSecurityManager().getUser()
		if not i:
			raise
	except:
		return None

	clusters = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/objectItems')('Folder')
	if not clusters:
		return None

	allowedClusters = list()
	for c in clusters:
		if i.has_role('View', c[1]):
			allowedClusters.append(c)

	return allowedClusters

def getStorage(self):
	if isAdmin(self):
		return self.restrictedTraverse(PLONE_ROOT + '/systems/storage/objectItems')('Folder')
	try:
		i = getSecurityManager().getUser()
		if not i:
			return None
	except:
		return None

	storage = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/objectItems')('Folder')
	if not storage:
		return None

	allowedStorage = list()
	for s in storage:
		if i.has_role('View', s[1]):
			allowedStorage.append(s)

	return allowedStorage

def createSystem(self, host, passwd):
	try:
		exists = self.restrictedTraverse(PLONE_ROOT +'/systems/storage/' + host)
		return 'Storage system \"' + host + '\" is already managed.'
	except:
		pass

	try:
		rc = RicciCommunicator(host)
		if rc is None:
			raise Exception, 'unknown error'
	except Exception, e:
		return 'Unable to establish a connection to the ricci agent on %s: %s' \
			% (host, str(e))

	try:
		if not rc.authed():
			rc.auth(passwd)
	except:
		return 'Unable to communicate with the ricci agent on \"' + host + '\" for authentication'

	try:
		i = rc.authed()
	except:
		return 'Unable to authenticate to the ricci agent on \"' + host + '\"'

	if i != True:
		return 'Authentication for storage system \"' + host + '\" failed'

#	rhost = rc.system_name()
#	if rhost and rhost != host and rhost[:9] != 'localhost' and rhost[:5] != '127.0':
#		host = str(rhost)

	try:
		exists = self.restrictedTraverse(PLONE_ROOT +'/systems/storage/' + host)
		return 'Storage system \"' + host + '\" is already managed.'
	except:
		pass

	try:
		ssystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/')
	except Exception, e:
		return 'Unable to create storage system %s: %s' % (host, str(e))

	try:
		ssystem.manage_addFolder(host, '__luci__:system')
		newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + host)
	except Exception, e:
		return 'Unable to create storage system %s: %s' % (host, str(e))

	try:
		newSystem.manage_acquiredPermissions([])
		newSystem.manage_role('View', ['Access contents information','View'])
	except Exception, e:
		return 'Unable to set permissions on storage system %s: %s' % (host, str(e))

	return None

def abortManageCluster(self, request):
	try:
		sessionData = request.SESSION.get('checkRet')
		nodeUnauth(sessionData['requestResults']['nodeList'])
	except:
		pass

def manageCluster(self, clusterName, nodeList):
	clusterName = str(clusterName)
	luci_log.debug_verbose('manageCluster for %s' % clusterName)

	try:
		clusters = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/')
		if not clusters:
			raise Exception, 'cannot find the cluster entry in the DB'
	except:
		nodeUnauth(nodeList)
		return 'Unable to create cluster \"' + clusterName + '\": the cluster directory is missing.'

	try:
		newCluster = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		if newCluster:
			nodeUnauth(nodeList)
			return 'A cluster named \"' + clusterName + '\" is already managed by Luci'
	except:
		pass

	try:
		clusters.manage_addFolder(clusterName, '__luci__:cluster')
		newCluster = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		if not newCluster:
			raise Exception, 'unable to find cluster folder for %s' % clusterName
	except Exception, e:
		nodeUnauth(nodeList)
		return 'Unable to create cluster %s: %s' % (clusterName, str(e))

	try:
		newCluster.manage_acquiredPermissions([])
		newCluster.manage_role('View', ['Access Contents Information','View'])
	except Exception, e:
		nodeUnauth(nodeList)
		try:
			clusters.manage_delObjects([clusterName])
		except:
			pass
		return 'Unable to set permissions on new cluster: %s: %s' % (clusterName, str(e))

	try:
		cluster_os = nodeList[0]['os']
		if not cluster_os:
			raise KeyError, 'Cluster OS is blank'
	except KeyError, e:
		luci_log.debug_verbose('Warning adding cluster %s: %s' \
			% (clusterName, str(e)))
		cluster_os = 'rhel5'

	try:
		newCluster.manage_addProperty('cluster_os', cluster_os, 'string')
	except:
		pass # we were unable to set the OS property string on this cluster

	for i in nodeList:
		#if 'ricci_host' in i:
		#	host = str(i['ricci_host'])
		#else:
		host = str(i['host'])

		try:
			newCluster.manage_addFolder(host, '__luci__:csystem:' + clusterName)
			newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName + '/' + host)
			if not newSystem:
				raise Exception, 'unable to create cluster system DB entry'
			newSystem.manage_acquiredPermissions([])
			newSystem.manage_role('View', [ 'Access contents information' , 'View' ])
		except Exception, e:
			nodeUnauth(nodeList)
			try:
				clusters.manage_delObjects([clusterName])
			except:
				pass
			return 'Unable to create cluster node %s for cluster %s: %s' \
				% (host, clusterName, str(e))

	try:
		ssystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/')
		if not ssystem:
			raise Exception, 'The storage DB entry is missing'
	except Exception, e:
		return 'Error adding storage node %s: %s' % (host, str(e))

	# Only add storage systems if the cluster and cluster node DB
	# objects were added successfully.
	for i in nodeList:
		#if 'ricci_host' in i:
		#	host = str(i['ricci_host'])
		#else:
		host = str(i['host'])

		try:
			# It's already there, as a storage system, no problem.
			exists = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + host)
			continue
		except: pass

		try:
			ssystem.manage_addFolder(host, '__luci__:system')
			newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + host)
			newSystem.manage_acquiredPermissions([])
			newSystem.manage_role('View', [ 'Access contents information' , 'View' ])
		except: pass

def createClusterSystems(self, clusterName, nodeList):
	try:
		clusterObj = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		if not clusterObj:
			raise Exception, 'cluster %s DB entry is missing' % clusterName
	except:
		nodeUnauth(nodeList)
		return 'No cluster named \"' + clusterName + '\" is managed by Luci'

	for i in nodeList:
		#if 'ricci_host' in i:
		#	host = str(i['ricci_host'])
		#else:
		host = str(i['host'])

		try:
			clusterObj.manage_addFolder(host, '__luci__:csystem:' + clusterName)
			newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName + '/' + host)
			if not newSystem:
				raise Exception, 'cluster node DB entry for %s disappeared from under us' % host
					
			newSystem.manage_acquiredPermissions([])
			newSystem.manage_role('View', [ 'Access contents information' , 'View' ])
		except Exception, e:
			nodeUnauth(nodeList)
			return 'Unable to create cluster node %s for cluster %s: %s' \
				% (host, clusterName, str(e))

	try:
		ssystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/')
		if not ssystem:
			raise Exception, 'storage DB entry is missing'
	except Exception, e:
		luci_log.debug_verbose('Error: adding storage DB node for %s: %s' \
			% (host, str(e)))
		return

	# Only add storage systems if the and cluster node DB
	# objects were added successfully.
	for i in nodeList:
		#if 'ricci_host' in i:
		#	host = str(i['ricci_host'])
		#else:
		host = str(i['host'])

		try:
			# It's already there, as a storage system, no problem.
			exists = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + host)
			continue
		except: pass

		try:
			ssystem.manage_addFolder(host, '__luci__:system')
			newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + host)
			newSystem.manage_acquiredPermissions([])
			newSystem.manage_role('View', [ 'Access contents information' , 'View' ])
		except: pass

def delSystem(self, systemName):
	try:
		ssystem = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/')
		if not ssystem:
			raise Exception, 'storage DB entry is missing'
	except Exception, e:
		return 'Unable to find storage system %s: %s' % (systemName, str(e))

	try:
		rc = RicciCommunicator(systemName)
		if rc is None:
			raise Exception, 'unknown error'
	except Exception, e:
		try:
			ssystem.manage_delObjects([systemName])
		except:
			return 'Unable to delete the storage system \"' + systemName + '\"'
		luci_log.debug_verbose('ricci error for %s: %s' % (systemName, str(e)))
		return

	# Only unauthenticate if the system isn't a member of
	# a managed cluster.
	cluster_info = rc.cluster_info()
	if not cluster_info[0]:
		try: rc.unauth()
		except: pass
	else:
		try:
			newSystem = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + cluster_info[0] + '/' + systemName)
		except:
			try: rc.unauth()
			except: pass

	try:
		ssystem.manage_delObjects([systemName])
	except Exception, e:
		return 'Unable to delete storage system %s: %s' \
			% (systemName, str(e))

def delCluster(self, clusterName):
	try:
		clusters = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/')
		if not clusters:
			raise
	except:
		return 'Unable to find cluster \"' + clusterName + '\"'

	err = delClusterSystems(self, clusterName)
	if err:
		return err

	try:
		clusters.manage_delObjects([clusterName])
	except:
		return 'Unable to delete cluster \"' + clusterName + '\"'

def delClusterSystem(self, cluster, systemName):
	try:
		if not self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + systemName):
			raise
	except:
		try:
			rc = RicciCommunicator(systemName)
			rc.unauth()
		except Exception, e:
			luci_log.debug_verbose('ricci error for %s: %s' \
				% (systemName, str(e)))

	try:
		cluster.manage_delObjects([systemName])
	except Exception, e:
		err_str = 'Error deleting cluster object %s: %s' % (systemName, str(e))
		luci_log.debug_verbose(err_str)
		return err_str

def delClusterSystems(self, clusterName):
	try:
		cluster = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + clusterName)
		if not cluster:
			raise Exception, 'cluster DB entry is missing'
		csystems = getClusterSystems(self, clusterName)
	except Exception, e:
		luci_log.debug_verbose('delCluSysterms: error for %s: %s' \
			% (clusterName, str(e)))
		return 'Unable to find any systems for cluster \"' + clusterName + '\"'

	errors = ''
	for i in csystems:
		err = delClusterSystem(self, cluster, i[0])
		if err:
			errors += 'Unable to delete the cluster system %s: %s\n' % (i[0], err)
	return errors

def getDefaultUser(self, request):
	try:
		user = request.form['userList']
	except:
		try:
			user = request['user']
		except:
			try:
				user = self.portal_membership.listMembers()[0].getUserName()
			except:
				user = None

	return user

def getUserPerms(self):
	perms = {}
	for i in self.portal_membership.listMembers():
		userName = i.getUserName()

		perms[userName] = {}
		perms[userName]['cluster'] = {}
		perms[userName]['storage'] = {}

		clusters = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/objectItems')('Folder')
		storage = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/objectItems')('Folder')

		for c in clusters:
			perms[userName]['cluster'][c[0]] = i.has_role('View', c[1])

		for s in storage:
			perms[userName]['storage'][s[0]] = i.has_role('View', s[1])

	return perms

# In case we want to give access to non-admin users in the future

def havePermCreateCluster(self):
	return isAdmin(self)

def havePermAddStorage(self):
	return isAdmin(self)

def havePermAddCluster(self):
	return isAdmin(self)

def havePermAddUser(self):
	return isAdmin(self)

def havePermDelUser(self):
	return isAdmin(self)

def havePermRemStorage(self):
	return isAdmin(self) 

def havePermRemCluster(self):
	return isAdmin(self) 

def havePermEditPerms(self):
	return isAdmin(self) 

def getClusterConfNodes(clusterConfDom):
	cur = clusterConfDom
	clusterNodes = list()

	for i in cur.childNodes:
		cur = i
		if i.nodeName == 'clusternodes':
			for i in cur.childNodes:
				if i.nodeName == 'clusternode':
					clusterNodes.append(i.getAttribute('name'))
			return clusterNodes
	return clusterNodes

def getSystems(self):
	storage = getStorage(self)
	clusters = getClusters(self)
	storageList = list()
	ret = [{}, [], {}]

	need_auth_hash = {}
	for i in storage:
		storageList.append(i[0])
		if testNodeFlag(i[1], CLUSTER_NODE_NEED_AUTH) != False:
			need_auth_hash[i[0]] = i[1]

	chash = {}
	for i in clusters:
		csystems = getClusterSystems(self, i[0])
		cslist = list()
		for c in csystems:
			if testNodeFlag(c[1], CLUSTER_NODE_NEED_AUTH) != False:
				need_auth_hash[c[0]] = c[1]
			cslist.append(c[0])
		chash[i[0]] = cslist

	ret[0] = chash
	ret[1] = storageList
	ret[2] = need_auth_hash
	return ret

def getClusterNode(self, nodename, clustername):
	try:
		cluster_node = self.restrictedTraverse(PLONE_ROOT + '/systems/cluster/' + str(clustername) + '/' + str(nodename))
		return cluster_node
	except:
		return None

def getStorageNode(self, nodename):
	try:
		storage_node = self.restrictedTraverse(PLONE_ROOT + '/systems/storage/' + '/' + str(nodename))
		return storage_node
	except:
		return None

def testNodeFlag(node, flag_mask):
	try:
		flags = node.getProperty('flags')
		return flags & flag_mask != 0
	except:
		pass
	return False

def setNodeFlag(node, flag_mask):
	try:
		flags = node.getProperty('flags')
		node.manage_changeProperties({ 'flags': flags | flag_mask })
	except:
		try:
			node.manage_addProperty('flags', flag_mask, 'int')
		except:
			pass

def delNodeFlag(node, flag_mask):
	try:
		flags = node.getProperty('flags')
		if flags & flag_mask != 0:
			node.manage_changeProperties({ 'flags': flags & ~flag_mask })
	except:
		pass
